import numpy as np
import copy
import pprint
import jsonschema
from yggdrasil.metaschema.datatypes import MetaschemaTypeError, YGG_MSG_HEAD
from yggdrasil.tests import YggTestClassInfo


class TestMetaschemaType(YggTestClassInfo):
    r"""Test class for MetaschemaType class."""

    _mod = 'MetaschemaType'
    _cls = 'MetaschemaType'
    _explicit = False

    def __init__(self, *args, **kwargs):
        super(TestMetaschemaType, self).__init__(*args, **kwargs)
        self._empty_msg = b''
        self._typedef = {}
        self._invalid_validate = [None]
        self._valid_encoded = []
        self._invalid_encoded = [{}]
        self._valid_decoded = ['nothing']
        self._invalid_decoded = []
        self._compatible_objects = []
        self._encode_type_kwargs = {}
        self._encode_data_kwargs = {}
        self._valid_normalize = [(None, None)]

    @property
    def mod(self):
        r"""str: Absolute name of module containing class to be tested."""
        return 'yggdrasil.metaschema.datatypes.%s' % self._mod

    @property
    def typedef(self):
        r"""dict: Type definition."""
        out = copy.deepcopy(self._typedef)
        out['type'] = self.import_cls.name
        return out

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for creating a class instance."""
        return self._typedef

    def assert_result_equal(self, x, y):
        r"""Assert that serialized/deserialized objects equal."""
        if isinstance(x, dict):
            if not isinstance(y, dict):  # pragma: debug
                raise AssertionError("Second variable is not a dictionary.")
            for k in x.keys():
                if k not in y:  # pragma: debug
                    print('x')
                    pprint.pprint(x)
                    print('y')
                    pprint.pprint(y)
                    raise AssertionError("Key '%s' not in second dictionary." % k)
                self.assert_result_equal(x[k], y[k])
            for k in y.keys():
                if k not in x:  # pragma: debug
                    print('x')
                    pprint.pprint(x)
                    print('y')
                    pprint.pprint(y)
                    raise AssertionError("Key '%s' not in first dictionary." % k)
        elif isinstance(x, (list, tuple)):
            if not isinstance(y, (list, tuple)):  # pragma: debug
                raise AssertionError("Second variable is not a list or tuple.")
            if len(x) != len(y):  # pragma: debug
                print('x')
                pprint.pprint(x)
                print('y')
                pprint.pprint(y)
                raise AssertionError("Sizes do not match. %d vs. %d"
                                     % (len(x), len(y)))
            for ix, iy in zip(x, y):
                self.assert_result_equal(ix, iy)
        elif isinstance(x, np.ndarray):
            np.testing.assert_array_equal(x, y)
        else:
            if isinstance(y, (dict, list, tuple, np.ndarray)):  # pragma: debug
                print('x')
                pprint.pprint(x)
                print('y')
                pprint.pprint(y)
                raise AssertionError("Compared objects are different types. "
                                     "%s vs. %s" % (type(x), type(y)))
            self.assert_equal(x, y)

    def test_validate(self):
        r"""Test validation."""
        if self._cls == 'MetaschemaType':
            for x in self._valid_decoded:
                self.assert_raises(NotImplementedError, self.import_cls.validate, x)
        else:
            for x in self._valid_decoded:
                self.assert_equal(self.import_cls.validate(x), True)
            for x in self._invalid_validate:
                self.assert_equal(self.import_cls.validate(x), False)
                self.assert_raises(BaseException, self.import_cls.validate,
                                   x, raise_errors=True)

    def test_normalize(self):
        r"""Test normalization."""
        for (x, y) in self._valid_normalize:
            z = self.import_cls.normalize(x)
            self.assert_result_equal(z, y)

    def test_fixed2base(self):
        r"""Test conversion of type definition from fixed type to the base."""
        if self._explicit:
            t1 = self.typedef
            x1 = self.import_cls.typedef_fixed2base(t1)
            t2 = copy.deepcopy(x1)
            t2['type'] = t1['type']
            x2 = self.import_cls.typedef_fixed2base(t2)
            self.assert_equal(x1, x2)
            y = self.import_cls.typedef_base2fixed(x1)
            self.assert_equal(y, self.typedef)

    def test_extract_typedef(self):
        r"""Test extract_typedef."""
        if len(self._valid_encoded) > 0:
            self.import_cls.extract_typedef(self._valid_encoded[0])

    def test_update_typedef(self):
        r"""Test update_typedef raises error on non-matching typename."""
        self.instance.update_typedef(**self.typedef)
        self.assert_raises(MetaschemaTypeError, self.instance.update_typedef,
                           type='invalid')
        if self._explicit:
            typedef_base = self.import_cls.typedef_fixed2base(self.typedef)
            self.instance.update_typedef(**typedef_base)

    def test_definition_schema(self):
        r"""Test definition schema."""
        s = self.import_cls.definition_schema()
        # jsonschema.Draft3Validator.check_schema(s)
        jsonschema.Draft4Validator.check_schema(s)

    def test_metadata_schema(self):
        r"""Test metadata schema."""
        s = self.import_cls.metadata_schema()
        # jsonschema.Draft3Validator.check_schema(s)
        jsonschema.Draft4Validator.check_schema(s)

    def test_encode_data(self):
        r"""Test encode/decode data & type."""
        if self._cls == 'MetaschemaType':
            for x in self._valid_decoded:
                self.assert_raises(NotImplementedError, self.import_cls.encode_type, x)
                self.assert_raises(NotImplementedError, self.import_cls.encode_data,
                                   x, self.typedef)
            self.assert_raises(NotImplementedError, self.import_cls.decode_data, None,
                               self.typedef)
        else:
            for x in self._valid_decoded:
                y = self.import_cls.encode_type(x, **self._encode_type_kwargs)
                z = self.import_cls.encode_data(x, y, **self._encode_data_kwargs)
                self.import_cls.encode_data_readable(x, None)
                self.import_cls.encode_data_readable(x, y)
                x2 = self.import_cls.decode_data(z, y)
                self.assert_result_equal(x2, x)
            if self._cls not in ['JSONNullMetaschemaType', 'AnyMetaschemaType']:
                self.assert_raises(MetaschemaTypeError,
                                   self.import_cls.encode_type, None)

    def test_check_encoded(self):
        r"""Test check_encoded."""
        # Test invalid for incorrect typedef
        if len(self._valid_encoded) > 0:
            self.assert_equal(self.import_cls.check_encoded(self._valid_encoded[0],
                                                            {}), False)
            self.assert_raises(BaseException, self.import_cls.check_encoded,
                               self._valid_encoded[0], {}, raise_errors=True)
        # Test valid
        for x in self._valid_encoded:
            assert(self.import_cls.check_encoded(x, self.typedef, raise_errors=True))
        # Test invalid
        for x in self._invalid_encoded:
            self.assert_equal(self.import_cls.check_encoded(x, self.typedef), False)
            self.assert_raises(BaseException, self.import_cls.check_encoded,
                               x, self.typedef, raise_errors=True)

    def test_check_decoded(self):
        r"""Test check_decoded."""
        # Not implemented for base class
        if self._cls == 'MetaschemaType':
            for x in self._valid_decoded:
                self.assert_raises(NotImplementedError, self.import_cls.check_decoded,
                                   x, self.typedef)
        else:
            # Test object alone
            if len(self._valid_decoded) > 0:
                x = self._valid_decoded[0]
                self.assert_equal(self.import_cls.check_decoded(x, None), True)
            # Test valid
            for x in self._valid_decoded:
                self.assert_equal(self.import_cls.check_decoded(x, self.typedef), True)
            # Test invalid with incorrect typedef
            for x in self._valid_decoded:
                self.assert_equal(self.import_cls.check_decoded(x, {}), False)
                self.assert_raises(BaseException, self.import_cls.check_decoded,
                                   x, {}, raise_errors=True)
            # Test invalid
            for x in (self._invalid_validate + self._invalid_decoded):
                self.assert_equal(self.import_cls.check_decoded(x, self.typedef), False)
                self.assert_raises(BaseException, self.import_cls.check_decoded,
                                   x, self.typedef, raise_errors=True)

    def test_encode_errors(self):
        r"""Test error on encode."""
        if self._cls == 'MetaschemaType':
            if self._invalid_validate:
                self.assert_raises(NotImplementedError, self.import_cls.encode,
                                   self._invalid_validate[0], self.typedef)
        else:
            if self._invalid_validate:
                self.assert_raises((ValueError,
                                    jsonschema.exceptions.ValidationError),
                                   self.import_cls.encode,
                                   self._invalid_validate[0], self.typedef)
            if self._valid_decoded:
                self.assert_raises(RuntimeError, self.import_cls.encode,
                                   self._valid_decoded[0], self.typedef,
                                   type='invalid')

    def test_decode_errors(self):
        r"""Test error on decode."""
        if self._invalid_encoded:
            self.assert_raises((ValueError, jsonschema.exceptions.ValidationError),
                               self.import_cls.decode,
                               self._invalid_encoded[0], self.typedef)

    def test_transform_type(self):
        r"""Test transform_type."""
        for x, y, typedef in self._compatible_objects:
            z = self.import_cls.transform_type(x, typedef)
            self.assert_result_equal(z, y)

    def test_serialize(self):
        r"""Test serialize/deserialize."""
        if self._cls == 'MetaschemaType':
            for x in self._valid_decoded:
                self.assert_raises(NotImplementedError, self.instance.serialize, x)
        else:
            for x in self._valid_decoded:
                msg = self.instance.serialize(x)
                y = self.instance.deserialize(msg)
                self.assert_result_equal(y[0], x)

    def test_serialize_error(self):
        r"""Test serialization errors."""
        if (self._cls != 'MetaschemaType') and (len(self._valid_decoded) > 0):
            self.assert_raises(RuntimeError, self.instance.serialize,
                               self._valid_decoded[0], data='something')

    def test_deserialize_error(self):
        r"""Test deserialization errors."""
        self.assert_raises(TypeError, self.instance.deserialize, self)
        self.assert_raises(ValueError, self.instance.deserialize,
                           b'invalid')
        if (self._cls != 'MetaschemaType') and (len(self._valid_decoded) > 0):
            self.assert_raises(ValueError, self.instance.deserialize,
                               self.instance.serialize(self._valid_decoded[0]),
                               metadata={'size': 0})
        
    def test_deserialize_empty(self):
        r"""Test call for empty string."""
        # Completely empty
        out = self.instance.deserialize(self._empty_msg)
        self.assert_result_equal(out[0], self.instance._empty_msg)
        self.assert_equal(out[1], dict(size=0, incomplete=False))
        # Empty metadata and message
        out = self.instance.deserialize((2 * YGG_MSG_HEAD) + self._empty_msg)
        self.assert_result_equal(out[0], self.instance._empty_msg)
        self.assert_equal(out[1], dict(size=0, incomplete=False))

    def test_deserialize_incomplete(self):
        r"""Test call for incomplete message."""
        if (self._cls != 'MetaschemaType') and (len(self._valid_decoded) > 0):
            out = self.instance.serialize(self._valid_decoded[0])
            obj, metadata = self.instance.deserialize(out[:-1])
            self.assert_equal(metadata['incomplete'], True)



import os
import re
import pprint
import unittest
import tempfile
from yggdrasil import platform
from yggdrasil.tests import scripts, assert_raises, assert_equal
import yggdrasil.drivers.tests.test_CompiledModelDriver as parent
from yggdrasil.drivers.CMakeModelDriver import (
    CMakeModelDriver, CMakeConfigure, CMakeBuilder)


_driver_installed = CMakeModelDriver.is_installed()


def test_CMakeConfigure():
    r"""Test CMakeConfigure."""
    src = scripts['c'][0]
    sourcedir = os.path.dirname(src)
    builddir = sourcedir
    # Test get_output_file
    out = CMakeConfigure.get_output_file(src, dont_build=True)
    assert_equal(out, builddir)
    out = CMakeConfigure.get_output_file(src, dont_build=True,
                                         builddir='.', working_dir=sourcedir)
    assert_equal(out, builddir)
    # Test get_flags
    out_A = CMakeConfigure.get_flags(dont_link=True)
    out_B = CMakeConfigure.get_flags(dont_link=True, outfile='.')
    assert_equal(out_A, out_B)


def test_CMakeBuilder():
    r"""Test CMakeBuilder."""
    src = scripts['c'][0]
    target = os.path.splitext(os.path.basename(src))[0]
    builddir = os.path.dirname(src)
    obj = os.path.splitext(src)[0] + '.obj'
    out = os.path.splitext(src)[0]
    if platform._is_win:  # pragma: windows
        out += '.exe'
    # Test get_output_file
    assert_equal(CMakeBuilder.get_output_file(obj), out)
    assert_equal(CMakeBuilder.get_output_file(obj, target='clean'), 'clean')
    assert_equal(CMakeBuilder.get_output_file(builddir, target=target), out)
    assert_raises(RuntimeError, CMakeBuilder.get_output_file, builddir)
    # Test get_flags
    out_A = CMakeBuilder.get_flags(target=target, working_dir=builddir)
    out_B = CMakeBuilder.get_flags(target=target,
                                   outfile=os.path.join('.', os.path.basename(out)))
    assert_equal(out_A, out_B)


@unittest.skipIf(not _driver_installed, "C Library not installed")
def test_create_include():
    r"""Test create_include."""
    target = 'target'
    tempdir = tempfile.gettempdir()
    fname_dll = os.path.join(tempdir, 'test.dll')
    fname_lib = os.path.join(tempdir, 'test.lib')
    for fname in [fname_dll, fname_lib]:
        with open(fname, 'w') as fd:
            fd.write('')
        assert(os.path.isfile(fname))
    testlist = [(['-DYGG'], [], ['ADD_DEFINITIONS(-DYGG)']),
                (['-Wall'], [], ['ADD_DEFINITIONS(-Wall)']),
                (['/nologo'], [], ['ADD_DEFINITIONS(/nologo)']),
                (['-Iinclude_dir'], [], ['INCLUDE_DIRECTORIES(include_dir)']),
                ([], ['-lm'], ['TARGET_LINK_LIBRARIES(%s -lm)' % target]),
                ([], ['-Llib_dir'], ['LINK_DIRECTORIES(lib_dir)']),
                ([], ['/LIBPATH:"lib_dir"'], ['LINK_DIRECTORIES(lib_dir)']),
                ([], ['m'], ['TARGET_LINK_LIBRARIES(%s m)' % target])]
    if CMakeConfigure.add_libraries:  # pragma: debug
        testlist += [([], [fname_dll], ['ADD_LIBRARY(test SHARED IMPORTED)']),
                     ([], [fname_lib], ['ADD_LIBRARY(test STATIC IMPORTED)'])]
    else:
        if platform._is_win:  # pragma: windows
            tempdir_cp = tempdir.replace('\\', re.escape('\\'))
        else:
            tempdir_cp = tempdir
        testlist += [([], [fname_dll], [('FIND_LIBRARY(TEST_LIBRARY NAMES %s '
                                         'test HINTS %s)')
                                        % (os.path.basename(fname_dll), tempdir_cp)]),
                     ([], [fname_lib], [('FIND_LIBRARY(TEST_LIBRARY NAMES %s '
                                         'test HINTS %s)')
                                        % (os.path.basename(fname_lib), tempdir_cp)])]
    from yggdrasil.drivers.CModelDriver import CModelDriver
    CModelDriver.compile_dependencies()
    CMakeModelDriver.compile_dependencies()
    for c, l, lines in testlist:
        out = CMakeConfigure.create_include(None, target, compile_flags=c,
                                            linker_flags=l, verbose=True)
        for x in lines:
            try:
                assert(x in out)
            except AssertionError:  # pragma: debug
                print("Could not find '%s':" % x)
                pprint.pprint(out)
                raise
    for fname in [fname_dll, fname_lib]:
        os.remove(fname)
    assert_raises(ValueError, CMakeConfigure.create_include,
                  None, target, compile_flags=['invalid'])
    assert_raises(ValueError, CMakeConfigure.create_include,
                  None, target, linker_flags=['-invalid'])
    assert_raises(ValueError, CMakeConfigure.create_include,
                  None, target, linker_flags=['/invalid'])


@unittest.skipIf(_driver_installed, "C Library installed")
def test_CMakeModelDriver_no_C_library():  # pragma: windows
    r"""Test CMakeModelDriver error when C library not installed."""
    assert_raises(RuntimeError, CMakeModelDriver, 'test', scripts['cmake'])


@unittest.skipIf(not _driver_installed, "C Library not installed")
def test_CMakeModelDriver_error_cmake():
    r"""Test CMakeModelDriver error for invalid cmake args."""
    makedir, target = os.path.split(scripts['cmake'])
    assert_raises(RuntimeError, CMakeModelDriver, 'test', target,
                  sourcedir=makedir, compiler_flags='-P',
                  target_language='c')


@unittest.skipIf(not _driver_installed, "C Library not installed")
def test_CMakeModelDriver_error_notarget():
    r"""Test CMakeModelDriver error for invalid target."""
    makedir, target = os.path.split(scripts['cmake'])
    assert_raises(RuntimeError, CMakeModelDriver, 'test', 'invalid',
                  sourcedir=makedir, target_language='c')


@unittest.skipIf(not _driver_installed, "C Library not installed")
def test_CMakeModelDriver_error_nofile():
    r"""Test CMakeModelDriver error for missing CMakeLists.txt."""
    assert_raises(RuntimeError, CMakeModelDriver, 'test', 'invalid',
                  target_language='c')


class TestCMakeModelParam(parent.TestCompiledModelParam):
    r"""Test parameters for CMakeModelDriver."""

    driver = 'CMakeModelDriver'
    
    def __init__(self, *args, **kwargs):
        super(TestCMakeModelParam, self).__init__(*args, **kwargs)
        self.attr_list += ['target', 'sourcedir', 'builddir']
        self.sourcedir, self.target = os.path.split(scripts['cmake'])
        self.builddir = os.path.join(self.sourcedir, 'build')
        self.args = [self.target]
        self._inst_kwargs['yml']['working_dir'] = self.sourcedir

    def test_sbdir(self):
        r"""Test that source/build directories set correctly."""
        assert_equal(self.instance.sourcedir, self.sourcedir)
        assert_equal(self.instance.builddir, self.builddir)
        

class TestCMakeModelDriverNoInit(TestCMakeModelParam,
                                 parent.TestCompiledModelDriverNoInit):
    r"""Test runner for CMakeModelDriver without init."""
    
    def test_sbdir(self):
        r"""Test that source/build directories set correctly."""
        pass
    
    
class TestCMakeModelDriverNoStart(TestCMakeModelParam,
                                  parent.TestCompiledModelDriverNoStart):
    r"""Test runner for CMakeModelDriver without start."""
    
    def __init__(self, *args, **kwargs):
        super(TestCMakeModelDriverNoStart, self).__init__(*args, **kwargs)
        # Version specifying sourcedir via working_dir
        self._inst_kwargs['yml']['working_dir'] = self.sourcedir
        # Relative paths
        self._inst_kwargs.update(sourcedir='.',
                                 builddir='build',
                                 compiler_flags=['-Wdev'])

    def test_call_compiler(self):
        r"""Test call_compiler without full path."""
        # self.instance.cleanup()
        self.import_cls.call_compiler(self.instance.source_files,
                                      builddir='build',
                                      working_dir=self.instance.working_dir,
                                      dont_build=True)
        self.import_cls.call_compiler(self.instance.source_files,
                                      out=self.instance.model_file,
                                      builddir='build',
                                      working_dir=self.instance.working_dir,
                                      overwrite=True)
        

class TestCMakeModelDriver(TestCMakeModelParam, parent.TestCompiledModelDriver):
    r"""Test runner for CMakeModelDriver."""

    def test_write_wrappers(self):
        r"""Test write_wrappers method with verbosity and existing
        include file."""
        self.instance.write_wrappers(verbose=True)



from yggdrasil.drivers.tests import test_ModelDriver as parent


class TestInterpretedModelParam(parent.TestModelParam):
    r"""Test parameters for basic InterpretedModelDriver class."""

    driver = 'InterpretedModelDriver'


class TestInterpretedModelDriver(TestInterpretedModelParam,
                                 parent.TestModelDriver):
    r"""Test runner for InterpretedModelDriver."""
    pass


class TestInterpretedModelDriverNoInit(TestInterpretedModelParam,
                                       parent.TestModelDriverNoInit):
    r"""Test runner for InterpretedModelDriver without init."""

    def test_executable_command(self):
        r"""Test error raise for invalid exec_type."""
        self.assert_raises(ValueError, self.import_cls.executable_command, [],
                           exec_type='invalid')

        
class TestInterpretedModelDriverNoStart(TestInterpretedModelParam,
                                        parent.TestModelDriverNoStart):
    r"""Test runner for InterpretedModelDriver without start."""
    pass



import os
import sys
import logging
import subprocess
import argparse
PY_MAJOR_VERSION = sys.version_info[0]
logger = logging.getLogger(__name__)
logger.setLevel(level=logging.INFO)


def update_argparser(parser=None):
    r"""Update argument parser with language specific arguments.

    Args:
        parser (argparse.ArgumentParser, optional): Existing argument parser
            that should be updated. Default to None and a new argument parser
            will be created.

    Returns:
        argparse.ArgumentParser: Argument parser with language specific arguments.

    """
    if parser is None:
        parser = argparse.ArgumentParser("Run Matlab installation script.")
    parser.add_argument('--user', action='store_true',
                        help=('Installation is being carried out in a user '
                              'directory.'))
    return parser


def install(args=None, as_user=None):
    r"""Attempt to install the MATLAB engine API for Python.

    Arguments:
        args (argparse.Namespace, optional): Arguments parsed from the
            command line. Default to None and is created from sys.argv.
        as_user (bool, optional): If True, the install will be called with
            --user for a local install. Defaults to None and args is
            checked for the flag '--user' to determine the value.

    Returns:
        bool: True if install succeded, False otherwise.

    """
    if args is None:
        args = update_argparser().parse_args()
    if as_user is None:
        as_user = args.user
    # Check to see if its already installed
    try:
        import matlab.engine as mtl_engine
        logger.info("MATLAB engine for python already installed.")
        del mtl_engine
        return True
    except ImportError:
        pass
    # Get location of matlab root
    mtl_id = '=MATLABROOT='
    cmd = "fprintf('" + mtl_id + "%s" + mtl_id + "', matlabroot); exit();"
    mtl_cmd = ['matlab', '-nodisplay', '-nosplash', '-nodesktop', '-nojvm',
               '-r', '%s' % cmd]
    try:
        mtl_proc = subprocess.check_output(mtl_cmd)
        if PY_MAJOR_VERSION == 3:
            mtl_proc = mtl_proc.decode("utf-8")
    except BaseException as e:
        logger.error("Error when calling MATLAB from the command line: %s"
                     % e)
        return False
    if mtl_id not in mtl_proc:
        logger.error("ID (%s) not in output from MATLAB process: %s"
                     % (mtl_id, mtl_proc))
        return False
    mtl_root = mtl_proc.split(mtl_id)[-2]
    # Install engine API
    mtl_setup = os.path.join(mtl_root, 'extern', 'engines', 'python')
    if not os.path.isdir(mtl_setup):
        logger.error("MATLAB directory dosn't exist: %s" % mtl_setup)
        return False
    blddir = os.path.join(os.path.expanduser('~'), 'matlab_python_api')
    if not os.path.isdir(blddir):
        os.mkdir(blddir)
    cmd = [sys.executable, 'setup.py',
           'build', '--build-base=%s' % blddir,
           'install']
    if as_user:
        cmd.append('--user')
    try:
        result = subprocess.check_output(cmd, cwd=mtl_setup)
        if PY_MAJOR_VERSION == 3:
            result = result.decode("utf-8")
        print(result)
    except subprocess.CalledProcessError as e:
        logger.error("Error calling install command for Matlab engine: %s" % e)
        return False
    return True


if __name__ == "__main__":
    out = install()
    if out:
        print("MATLAB engine installed.")
    else:
        raise Exception("Failed to install MATLAB engine.")



import sys
# Import Python module containing model calculation
from shoot import calc_shoot_mass


# Create input/output channels
if len(sys.argv) != 6:
    raise RuntimeError("shoot: 5 input files and 1 output files are required.")
ShootGrowthRate = open(sys.argv[1], 'r')
InitShootMass = open(sys.argv[2], 'r')
TimeStep = open(sys.argv[3], 'r')
NextRootMass = open(sys.argv[4], 'r')
NextShootMass = open(sys.argv[5], 'w')

# Read shoot growth rate
flag = False
while True:
    line = ShootGrowthRate.readline()
    if len(line) == 0:
        break
    if line.startswith('#'):
        continue
    r_s = float(line)
    flag = True
    break
if not flag:
    raise RuntimeError('shoot: Error reading shoot growth rate.')
ShootGrowthRate.close()
print('shoot: Read shoot growth rate: %f' % r_s)

# Read initial shoot mass
flag = False
while True:
    line = InitShootMass.readline()
    if len(line) == 0:
        break
    if line.startswith('#'):
        continue
    S_t = float(line)
    flag = True
    break
if not flag:
    raise RuntimeError('shoot: Error reading initial shoot mass.')
InitShootMass.close()
print('shoot: Read initial shoot mass: %f' % S_t)

# Read inital root mass
flag = False
while True:
    line = NextRootMass.readline()
    if len(line) == 0:
        break
    if line.startswith('#'):
        continue
    R_t = float(line) / 1000.0  # Conversion to kg
    flag = True
    break
if not flag:
    raise RuntimeError('shoot: Error reading initial shoot mass.')
print('shoot: Read initial root mass: %f' % R_t)

# Write initial shoot mass
NextShootMass.write('# shoot_mass\n# %%lf\n%lf\n' % S_t)

# Keep advancing until there arn't any new input times
i = 0
while True:
    
    # Read the time step
    buff = TimeStep.readline()
    if buff.startswith('#'):
        continue
    if len(buff) == 0:
        print('shoot: No more time steps.')
        break
    dt = float(buff) / 24.0  # Conversion from hours to days
    print('shoot: Read next time step: %f' % dt)

    # Read the next root mass
    buff = NextRootMass.readline()
    if buff.startswith('#') or (len(buff) == 0):
        # This raises an error because there must be a root mass for each time step
        raise RuntimeError('shoot: Error reading root mass for timestep %d.' % (i + 1))
    R_tp1 = float(buff) / 1000.0  # Conversion to kg
    print('shoot: Read next root mass: %f' % R_tp1)

    # Calculate shoot mass
    S_tp1 = calc_shoot_mass(r_s, dt, S_t, R_t, R_tp1)
    print('shoot: Calculated next shoot mass: %f' % S_tp1)

    # Output shoot mass
    NextShootMass.write('%lf\n' % S_tp1)

    # Advance masses to next timestep
    S_t = S_tp1
    R_t = R_tp1
    i += 1
    # break  # Only for timing of a single loop



import os
import copy
import shutil
import tempfile
import unittest
from yggdrasil.tests import assert_raises, assert_equal
from yggdrasil.metaschema.datatypes.tests import (
    test_PlyMetaschemaType as parent)
from yggdrasil.metaschema.datatypes import ObjMetaschemaType
from yggdrasil.drivers.LPyModelDriver import LPyModelDriver


old_value = parent._test_value
_test_value = {'vertices': [], 'faces': [], 'lines': [],
               'params': [{'u': 0.0, 'v': 0.0, 'w': 1.0},
                          {'u': 0.0, 'v': 0.0}],
               'normals': [{'i': 0.0, 'j': 0.0, 'k': 0.0},
                           {'i': 1.0, 'j': 1.0, 'k': 1.0}],
               'texcoords': [{'u': 0.0, 'v': 0.0, 'w': 0.0},
                             {'u': 1.0, 'v': 1.0},
                             {'u': 1.0}],
               'points': [[0, 2]],
               'curves': [{'starting_param': 0.0, 'ending_param': 1.0,
                           'vertex_indices': [0, 1]}],
               'curve2Ds': [[0, 1]],
               'surfaces': [{
                   'starting_param_u': 0.0, 'ending_param_u': 1.0,
                   'starting_param_v': 0.0, 'ending_param_v': 1.0,
                   'vertex_indices': [{'vertex_index': 0,
                                       'normal_index': 0},
                                      {'vertex_index': 1,
                                       'texcoord_index': 1,
                                       'normal_index': 1}]}]}
_test_value['material'] = old_value['material']
_test_value['vertices'] = copy.deepcopy(old_value['vertices'])
_test_value['vertices'][0]['w'] = 1.0
for f in old_value['faces']:
    new = [{'vertex_index': x, 'texcoord_index': 0} for x in f['vertex_index']]
    _test_value['faces'].append(new)
for fv in _test_value['faces'][0]:
    fv['normal_index'] = 0
_test_value['faces'][1][0].pop('texcoord_index')
_test_value['faces'][1][0]['normal_index'] = 0
for e in old_value['edges']:
    new = [{'vertex_index': e['vertex%d' % x]} for x in [1, 2]]
    _test_value['lines'].append(new)
_test_value_simple = {'vertices': copy.deepcopy(_test_value['vertices']),
                      'normals': copy.deepcopy(_test_value['normals']),
                      'texcoords': copy.deepcopy(_test_value['texcoords']),
                      'faces': [[{'vertex_index': 0, 'normal_index': 0},
                                 {'vertex_index': 1, 'normal_index': 0},
                                 {'vertex_index': 2, 'normal_index': 0}],
                                [{'vertex_index': 0, 'normal_index': 1},
                                 {'vertex_index': 2, 'normal_index': 1},
                                 {'vertex_index': 3, 'normal_index': 1}]]}
for f in _test_value_simple['faces']:
    for fv in f:
        fv['texcoord_index'] = 0
for v in _test_value_simple['vertices']:
    v.pop('w', None)
for t in _test_value_simple['texcoords']:
    t.pop('w', None)
    t.setdefault('v', 0.0)


def test_create_schema():
    r"""Test create_schema."""
    assert_raises(RuntimeError, ObjMetaschemaType.create_schema, overwrite=False)
    temp = os.path.join(tempfile.gettempdir(), 'temp_schema')
    old_schema = ObjMetaschemaType.get_schema()
    try:
        shutil.move(ObjMetaschemaType._schema_file, temp)
        new_schema = ObjMetaschemaType.get_schema()
        assert_equal(old_schema, new_schema)
    except BaseException:  # pragma: debug
        shutil.move(temp, ObjMetaschemaType._schema_file)
        raise
    shutil.move(temp, ObjMetaschemaType._schema_file)


class TestObjDict(parent.TestPlyDict):
    r"""Test for ObjDict class."""
    
    _mod = 'ObjMetaschemaType'
    _cls = 'ObjDict'
    _simple_test = _test_value_simple

    @property
    def inst_kwargs(self):
        r"""dict: Keyword arguments for creating a class instance."""
        return _test_value

    def test_apply_scalar_map(self):
        r"""Test applying a scalar colormap."""
        super(TestObjDict, self).test_apply_scalar_map(_as_obj=True)

    @unittest.skipIf(not LPyModelDriver.is_installed(), "LPy library not installed.")
    def test_to_from_scene(self):  # pragma: lpy
        r"""Test conversion to/from PlantGL scene."""
        super(TestObjDict, self).test_to_from_scene(_as_obj=True)


class TestObjMetaschemaType(parent.TestPlyMetaschemaType):
    r"""Test class for ObjMetaschemaType class with float."""

    _mod = 'ObjMetaschemaType'
    _cls = 'ObjMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestObjMetaschemaType, self).__init__(*args, **kwargs)
        self._value = _test_value
        self._fulldef = {'type': self.import_cls.name}
        self._typedef = {'type': self.import_cls.name}
        self._valid_encoded = [self._fulldef]
        self._valid_decoded = [self._value,
                               ObjMetaschemaType.ObjDict(**_test_value),
                               {'vertices': self._value['vertices'],
                                'faces': [[{'vertex_index': 0},
                                           {'vertex_index': 1},
                                           {'vertex_index': 2}]]}]
        self._invalid_encoded = [{}]
        self._invalid_decoded = [{'vertices': [{k: 0.0 for k in 'xyz'}],
                                  'faces': [[{'vertex_index': 0},
                                             {'vertex_index': 1},
                                             {'vertex_index': 2}]]},
                                 {'vertices': [], 'faces': None,
                                  'lines': [[None]],
                                  'surfaces': [{'vertex_indices': [[]]}]},
                                 {'vertices': self._value['vertices'],
                                  'texcoords': self._value['texcoords'],
                                  'normals': self._value['normals'],
                                  'faces': [[{'vertex_index': 0,
                                              'texcoord_index': 100,
                                              'normal_index': 100},
                                             {'vertex_index': 1},
                                             {'vertex_index': 2}]]},
                                 None]
        self._compatible_objects = [(self._value, self._value, None)]



from yggdrasil.serialize.SerializeBase import SerializeBase


class FunctionalSerialize(SerializeBase):
    r"""Class for serializing/deserializing a Python object into/from a bytes
    message using defined functions.

    Args:
        encoded_datatype (schema, optional): JSON schema describing the type
            that serialized objects should conform to. Defaults to the class
            attribute default_encoded_datatype. If either func_serialize or
            func_deserialize are not provided, this needs to be specified in
            order to serialize non-bytes objects.
        func_serialize (func, optional): Callable object that takes Python
            objects as input and returns a representation that conforms to
            encoded_datatype. Defaults to None and the default serialization
            for encoded_datatype will be used.
        func_deserialize (func, optional): Callable object that takes objects
            of a type that conforms to encoded_datatype and returns a
            deserialized Python object. Defaults to None and the default
            deserialization for encoded_datatype will be used.
        **kwargs: Additional keyword args are passed to the parent class's
            constructor.

    """

    _seritype = 'functional'
    _schema_subtype_description = ('Serializer that uses provied function to '
                                   'serialize messages.')
    _schema_requried = []
    _schema_properties = {
        'encoded_datatype': {'type': 'schema'},
        'func_serialize': {'type': 'function'},
        'func_deserialize': {'type': 'function'}}
    func_serialize = None
    func_deserialize = None
    
    def __init__(self, **kwargs):
        if isinstance(kwargs.get('func_serialize', None), SerializeBase):
            kwargs['func_serialize'] = kwargs['func_serialize'].func_serialize
        if isinstance(kwargs.get('func_deserialize', None), SerializeBase):
            kwargs['func_deserialize'] = kwargs['func_deserialize'].func_deserialize
        super(FunctionalSerialize, self).__init__(**kwargs)

    # @property
    # def base_class(self):
    #     r"""DefaultSerialize: Default version of serialization."""
    #     if getattr(self, '_base_class', None) is None:
    #         self._base_class = DefaultSerialize(datatype=self.typedef,
    #                                             **self.serializer_info)
    #     return self._base_class

    # TODO: In some cases this should be the object typedef
    # @property
    # def typedef(self):
    #     r"""dict: Type definition."""
    #     return self.encoded_typedef

    @property
    def serializer_info(self):
        r"""dict: Serializer info."""
        raise RuntimeError("Cannot define serializer information for user "
                           + "supplied functions.")

    @property
    def empty_msg(self):
        r"""obj: Object indicating empty message."""
        return self.encoded_datatype._empty_msg



import pprint
# Import classes for input/output channels
from yggdrasil.interface.YggInterface import (
    YggPlyInput, YggPlyOutput)

# Initialize input/output channels
in_channel = YggPlyInput('inputA')
out_channel = YggPlyOutput('outputA')

# Loop until there is no longer input or the queues are closed
while True:

    # Receive input from input channel
    # If there is an error, the flag will be False
    flag, ply = in_channel.recv()
    if not flag:
        print("Model A: No more input.")
        break

    # Print received message
    print('Model A: (%d verts, %d faces)' % (ply.nvert, ply.nface))
    pprint.pprint(ply)

    # Send output to output channel
    # If there is an error, the flag will be False
    flag = out_channel.send(ply)
    if not flag:
        raise RuntimeError("Model A: Error sending output.")



import numpy as np
import pandas as pd
from yggdrasil.metaschema.datatypes.ContainerMetaschemaType import (
    ContainerMetaschemaType)


class JSONArrayMetaschemaType(ContainerMetaschemaType):
    r"""Type associated with a set of subtypes.

    Developer Notes:
        Support for dynamic arrays in C/C++ is still under development.

    """

    name = 'array'
    description = 'A container of ordered values.'
    properties = ['items']
    metadata_properties = ['items']
    extract_properties = ['items']
    python_types = (list, tuple, np.ndarray, pd.DataFrame)
    _replaces_existing = True

    _container_type = list
    _json_type = 'array'
    _json_property = 'items'
    _empty_msg = []

    @classmethod
    def validate(cls, obj, raise_errors=False):
        r"""Validate an object to check if it could be of this type.

        Args:
            obj (object): Object to validate.
            raise_errors (bool, optional): If True, errors will be raised when
                the object fails to be validated. Defaults to False.

        Returns:
            bool: True if the object could be of this type, False otherwise.

        """
        out = super(JSONArrayMetaschemaType, cls).validate(
            obj, raise_errors=raise_errors)
        if out and isinstance(obj, np.ndarray):
            out = (len(obj.dtype) > 0)
            if (not out) and raise_errors:
                raise ValueError("Array dosn't have a structured data type.")
        return out

    @classmethod
    def normalize(cls, obj):
        r"""Normalize an object, if possible, to conform to this type.

        Args:
            obj (object): Object to normalize.

        Returns:
            object: Normalized object.

        """
        if isinstance(obj, str):
            obj = [v.strip() for v in obj.split(',')]
        elif isinstance(obj, tuple):
            obj = list(obj)
        return obj

    @classmethod
    def coerce_type(cls, obj, typedef=None, key_order=None, **kwargs):
        r"""Coerce objects of specific types to match the data type.

        Args:
            obj (object): Object to be coerced.
            typedef (dict, optional): Type defintion that object should be
                coerced to. Defaults to None.
            key_order (list, optional): Order that keys from a dictionary should
                be used to compose an array. Defaults to None.
            **kwargs: Additional keyword arguments are metadata entries that may
                aid in coercing the type.

        Returns:
            object: Coerced object.

        Raises:
            RuntimeError: If obj is a dictionary, but key_order is not provided.

        """
        from yggdrasil.serialize import pandas2list, numpy2list, dict2list
        if isinstance(obj, pd.DataFrame):
            obj = pandas2list(obj)
        elif isinstance(obj, np.ndarray) and (len(obj.dtype) > 0):
            obj = numpy2list(obj)
        elif isinstance(obj, dict):
            if (key_order is not None) or (len(obj) == 1):
                obj = dict2list(obj, order=key_order)
            else:
                obj = [obj]
        elif isinstance(typedef, dict) and (len(typedef.get('items', [])) == 1):
            typedef_validated = kwargs.get('typedef_validated', False)
            if cls.check_decoded([obj], typedef,
                                 typedef_validated=typedef_validated):
                obj = [obj]
        return obj

    @classmethod
    def _iterate(cls, container):
        r"""Iterate over the contents of the container. Each element returned
        should be a tuple including an index and a value.

        Args:
            container (obj): Object to be iterated over.

        Returns:
            iterator: Iterator over elements in the container.

        """
        for k, v in enumerate(container):
            yield (k, v)

    @classmethod
    def _assign(cls, container, index, value):
        r"""Assign an element in the container to the specified value.

        Args:
            container (obj): Object that element will be assigned to.
            index (obj): Index in the container object where element will be
                assigned.
            value (obj): Value that will be assigned to the element in the
                container object.

        """
        if len(container) > index:
            container[index] = value
        elif len(container) == index:
            container.append(value)
        else:
            raise RuntimeError("The container has %s elements and the index is %s"
                               % (len(container), index))

    @classmethod
    def _has_element(cls, container, index):
        r"""Check to see if an index is in the container.

        Args:
            container (obj): Object that should be checked for index.
            index (obj): Index that should be checked for.

        Returns:
            bool: True if the index is in the container.

        """
        return (len(container) > index)

    @classmethod
    def _get_element(cls, container, index, default):
        r"""Get an element from the container if it exists, otherwise return
        the default.

        Args:
            container (obj): Object that should be returned from.
            index (obj): Index of element that should be returned.
            default (obj): Default that should be returned if the index is not
                in the container.

        Returns:
            object: Container contents at specified element.

        """
        if isinstance(container, dict):
            assert('type' in container)
            return container
        return super(JSONArrayMetaschemaType, cls)._get_element(
            container, index, default)



#!/usr/bin/python
import os
import sys
import copy
import logging
import traceback
import subprocess
import argparse


logger = logging.getLogger(__name__)


def ygginfo():
    r"""Print information about yggdrasil installation."""
    from yggdrasil import __version__, tools, config, backwards, platform
    from yggdrasil.components import import_component
    lang_list = tools.get_installed_lang()
    prefix = '    '
    curr_prefix = ''
    vardict = [
        ('Location', os.path.dirname(__file__)),
        ('Version', __version__),
        ('Languages', ', '.join(lang_list)),
        ('Communication Mechanisms', ', '.join(tools.get_installed_comm())),
        ('Default Comm Mechanism', tools.get_default_comm()),
        ('Config File', config.usr_config_file)]
    parser = argparse.ArgumentParser(
        description='Display information about the current yggdrasil installation.')
    parser.add_argument('--no-languages', action='store_true',
                        dest='no_languages',
                        help='Don\'t print information about individual languages.')
    parser.add_argument('--verbose', action='store_true',
                        help='Increase the verbosity of the printed information.')
    args = parser.parse_args()
    try:
        # Add language information
        if not args.no_languages:
            # Install languages
            vardict.append(('Installed Languages:', ''))
            curr_prefix += prefix
            for lang in sorted(lang_list):
                drv = import_component('model', lang)
                vardict.append((curr_prefix + '%s:' % lang.upper(), ''))
                curr_prefix += prefix
                if lang == 'executable':
                    vardict.append((curr_prefix + 'Location', ''))
                else:
                    exec_name = drv.language_executable()
                    if not os.path.isabs(exec_name):
                        exec_name = tools.which(exec_name)
                    vardict.append((curr_prefix + 'Location', exec_name))
                vardict.append((curr_prefix + 'Version',
                                drv.language_version()))
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
            curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
            # Not installed languages
            vardict.append(("Languages Not Installed:", ''))
            curr_prefix += prefix
            for lang in tools.get_supported_lang():
                if lang in lang_list:
                    continue
                drv = import_component('model', lang)
                vardict.append((curr_prefix + '%s:' % lang.upper(), ''))
                curr_prefix += prefix
                vardict.append((curr_prefix + "Language Installed",
                                drv.is_language_installed()))
                vardict.append((curr_prefix + "Dependencies Installed",
                                drv.are_dependencies_installed()))
                vardict.append((curr_prefix + "Interface Installed",
                                drv.is_interface_installed()))
                vardict.append((curr_prefix + "Comm Installed",
                                drv.is_comm_installed()))
                vardict.append((curr_prefix + "Configured",
                                drv.is_configured()))
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
            curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
        # Add verbose information
        if args.verbose:
            # Conda info
            if os.environ.get('CONDA_PREFIX', ''):
                out = backwards.as_str(subprocess.check_output(
                    ['conda', 'info'])).strip()
                curr_prefix += prefix
                vardict.append((curr_prefix + 'Conda Info:', "\n%s%s"
                                % (curr_prefix + prefix,
                                   ("\n" + curr_prefix + prefix).join(
                                       out.splitlines(False)))))
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
            # R and reticulate info
            Rdrv = import_component("model", "R")
            if Rdrv.is_installed():
                env_reticulate = copy.deepcopy(os.environ)
                env_reticulate['RETICULATE_PYTHON'] = sys.executable
                # Stack size
                out = Rdrv.run_executable(["-e", "Cstack_info()"]).strip()
                vardict.append((curr_prefix + "R Cstack_info:", "\n%s%s"
                                % (curr_prefix + prefix,
                                   ("\n" + curr_prefix + prefix).join(
                                       out.splitlines(False)))))
                # Compilation tools
                interp = 'R'.join(Rdrv.get_interpreter().rsplit('Rscript', 1))
                vardict.append((curr_prefix + "R C Compiler:", ""))
                curr_prefix += prefix
                for x in ['CC', 'CFLAGS', 'CXX', 'CXXFLAGS']:
                    out = backwards.as_str(subprocess.check_output(
                        [interp, 'CMD', 'config', x])).strip()
                    vardict.append((curr_prefix + x, "%s"
                                    % ("\n" + curr_prefix + prefix).join(
                                        out.splitlines(False))))
                curr_prefix = curr_prefix.rsplit(prefix, 1)[0]
                # Session info
                out = Rdrv.run_executable(["-e", "sessionInfo()"]).strip()
                vardict.append((curr_prefix + "R sessionInfo:", "\n%s%s"
                                % (curr_prefix + prefix,
                                   ("\n" + curr_prefix + prefix).join(
                                       out.splitlines(False)))))
                # Reticulate conda_list
                if os.environ.get('CONDA_PREFIX', ''):
                    out = Rdrv.run_executable(
                        ["-e", ("library(reticulate); "
                                "reticulate::conda_list()")],
                        env=env_reticulate).strip()
                    vardict.append((curr_prefix + "R reticulate::conda_list():",
                                    "\n%s%s" % (curr_prefix + prefix,
                                                ("\n" + curr_prefix + prefix).join(
                                                    out.splitlines(False)))))
                # Windows python versions
                if platform._is_win:  # pragma: windows
                    out = Rdrv.run_executable(
                        ["-e", ("library(reticulate); "
                                "reticulate::py_versions_windows()")],
                        env=env_reticulate).strip()
                    vardict.append((curr_prefix
                                    + "R reticulate::py_versions_windows():",
                                    "\n%s%s" % (curr_prefix + prefix,
                                                ("\n" + curr_prefix + prefix).join(
                                                    out.splitlines(False)))))
                # conda_binary
                if platform._is_win:  # pragma: windows
                    out = Rdrv.run_executable(
                        ["-e", ("library(reticulate); "
                                "conda <- reticulate:::conda_binary(\"auto\"); "
                                "system(paste(conda, \"info --json\"))")],
                        env=env_reticulate).strip()
                    vardict.append((curr_prefix
                                    + "R reticulate::py_versions_windows():",
                                    "\n%s%s" % (curr_prefix + prefix,
                                                ("\n" + curr_prefix + prefix).join(
                                                    out.splitlines(False)))))
                # Reticulate py_config
                out = Rdrv.run_executable(["-e", ("library(reticulate); "
                                                  "reticulate::py_config()")],
                                          env=env_reticulate).strip()
                vardict.append((curr_prefix + "R reticulate::py_config():",
                                "\n%s%s" % (curr_prefix + prefix,
                                            ("\n" + curr_prefix + prefix).join(
                                                out.splitlines(False)))))
    finally:
        # Print things
        max_len = max(len(x[0]) for x in vardict)
        lines = []
        line_format = '%-' + str(max_len) + 's' + prefix + '%s'
        for k, v in vardict:
            lines.append(line_format % (k, v))
        logger.info("yggdrasil info:\n%s" % '\n'.join(lines))


def yggrun():
    r"""Start a run."""
    from yggdrasil import runner
    parser = argparse.ArgumentParser(description='Run an integration.')
    parser.add_argument('yamlfile', nargs='+',
                        help='One or more yaml specification files.')
    args = parser.parse_args()
    prog = sys.argv[0].split(os.path.sep)[-1]
    yggRunner = runner.get_runner(args.yamlfile, ygg_debug_prefix=prog)
    try:
        yggRunner.run()
        yggRunner.debug("runner returns, exiting")
    except Exception as ex:
        yggRunner.pprint("yggrun exception: %s" % type(ex))
        print(traceback.format_exc())
    print('')


def yggcc():
    r"""Compile C/C++ program."""
    from yggdrasil.drivers import CModelDriver
    parser = argparse.ArgumentParser(description='Compile a C/C++ program.')
    parser.add_argument('source', nargs='+',
                        help='One or more source files.')
    args = parser.parse_args()
    out = CModelDriver.CModelDriver.call_compile(args.source)
    print("executable: %s" % out)


def cc_flags():
    r"""Get the compiler flags necessary for including the interface
    library in a C or C++ program.

    Returns:
        list: The necessary compiler flags and preprocessor definitions.

    """
    from yggdrasil.drivers import CModelDriver
    print(' '.join(CModelDriver.CModelDriver.get_compiler_flags()))


def ld_flags():
    r"""Get the linker flags necessary for calling functions/classes from
    the interface library in a C or C++ program.

    Returns:
        list: The necessary library linking flags.

    """
    from yggdrasil.drivers import CModelDriver
    print(' '.join(CModelDriver.CModelDriver.get_linker_flags()))


def rebuild_c_api():
    r"""Rebuild the C/C++ API."""
    from yggdrasil.drivers import CModelDriver, CPPModelDriver
    if CModelDriver.CModelDriver.is_installed():
        CModelDriver.CModelDriver.compile_dependencies(overwrite=True)
        # TODO: Check that this compiles library correctly
        CPPModelDriver.CPPModelDriver.compile_dependencies(overwrite=True)
    else:
        raise Exception("The libraries necessary for running models written in "
                        "C/C++ could not be located.")

    
def regen_metaschema():
    r"""Regenerate the yggdrasil metaschema."""
    from yggdrasil import metaschema
    if os.path.isfile(metaschema._metaschema_fname):
        os.remove(metaschema._metaschema_fname)
    metaschema._metaschema = None
    metaschema._validator = None
    metaschema.get_metaschema()
    

def regen_schema():
    r"""Regenerate the yggdrasil schema."""
    from yggdrasil import schema
    if os.path.isfile(schema._schema_fname):
        os.remove(schema._schema_fname)
    schema.clear_schema()
    schema.init_schema()


def validate_yaml():
    r"""Validate a set of or or more YAMLs defining an integration."""
    from yggdrasil import yamlfile
    parser = argparse.ArgumentParser(
        description='Validate a set of YAML specification files for an integration.')
    parser.add_argument('yamlfile', nargs='+',
                        help='One or more YAML specification files.')
    args = parser.parse_args()
    yamlfile.parse_yaml(args.yamlfile)
    logger.info("Validation succesful.")


def update_config():
    r"""Update the user config file for yggdrasil."""
    from yggdrasil import config, tools
    from yggdrasil.components import import_component
    parser = argparse.ArgumentParser(
        description='Update the user config file.')
    parser.add_argument('--show-file', action='store_true',
                        help='Print the path to the config file without updating it.')
    parser.add_argument('--remove-file', action='store_true',
                        help='Remove the existing config file and return.')
    parser.add_argument('--overwrite', action='store_true',
                        help='Overwrite the existing file.')
    parser.add_argument('--languages', nargs='+',
                        default=tools.get_supported_lang(),
                        help=('One or more languages that should be'
                              'configured.'))
    parser.add_argument('--disable-languages', nargs='+',
                        default=[], dest='disable_languages',
                        help='One or more languages that should be disabled.')
    parser.add_argument('--enable-languages', nargs='+', default=[],
                        help='One or more languages that should be enabled.')
    args = parser.parse_args()
    if args.show_file:
        print('Config file located here: %s' % config.usr_config_file)
    if args.remove_file and os.path.isfile(config.usr_config_file):
        os.remove(config.usr_config_file)
    if args.show_file or args.remove_file:
        return
    drv = [import_component('model', l) for l in args.languages]
    config.update_language_config(drv, overwrite=args.overwrite, verbose=True,
                                  disable_languages=args.disable_languages,
                                  enable_languages=args.enable_languages)


def yggtime_comm():
    r"""Plot timing statistics comparing the different communication mechanisms."""
    from yggdrasil import timing
    timing.plot_scalings(compare='commtype')


def yggtime_lang():
    r"""Plot timing statistics comparing the different languages."""
    from yggdrasil import timing
    timing.plot_scalings(compare='language')


def yggtime_os():
    r"""Plot timing statistics comparing the different operating systems."""
    from yggdrasil import timing
    timing.plot_scalings(compare='platform')


def yggtime_py():
    r"""Plot timing statistics comparing the different versions of Python."""
    from yggdrasil import timing
    timing.plot_scalings(compare='python')


def yggtime_paper():
    r"""Create plots for timing."""
    from yggdrasil import timing
    _lang_list = timing._lang_list
    _lang_list_nomatlab = copy.deepcopy(_lang_list)
    _lang_list_nomatlab.remove('matlab')
    timing.plot_scalings(compare='platform', python_ver='2.7')
    # All plots on Linux, no matlab
    timing.plot_scalings(compare='comm_type', platform='Linux', python_ver='2.7')
    timing.plot_scalings(compare='python_ver', platform='Linux')
    timing.plot_scalings(compare='language', platform='Linux', python_ver='2.7',
                         compare_values=_lang_list_nomatlab)
    # Language comparision on MacOS, with matlab
    timing.plot_scalings(compare='language', platform='MacOS', python_ver='2.7',
                         compare_values=_lang_list)


def ygginstall():
    r"""Call installation script."""
    from yggdrasil.languages import install_languages
    parser = install_languages.update_argparser()
    args = parser.parse_args()
    if (len(args.language) == 0) or ('all' in args.language):
        install_languages.install_all_languages(args=args)
    else:
        for x in args.language:
            install_languages.install_language(x, args=args)


if __name__ == '__main__':
    yggrun()
    sys.exit(0)



import platform as sys_platform
from yggdrasil import tools, platform
from yggdrasil.drivers.ModelDriver import ModelDriver


# Version flags for windows appear to cause hang so just use
# platform module for now
# if platform._is_win:  # pragma: windows
#     _os_version_flags = ['winver']
# else:
#     _os_version_flags = ['uname', '-r']


class ExecutableModelDriver(ModelDriver):
    r"""Class for running executable based models."""

    language = 'executable'
    # version_flags = _os_version_flags
    _schema_subtype_description = ('Model is an executable.')

    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration including things like platform dependent properties and
        checking environment variables for default settings.
        """
        if platform._is_win:  # pragma: windows
            cls.language_ext = '.exe'
        ModelDriver.before_registration(cls)
        
    @classmethod
    def language_version(cls, version_flags=None, **kwargs):
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to cls.run_executable.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        return sys_platform.platform()
        
    @classmethod
    def executable_command(cls, args, unused_kwargs=None, **kwargs):
        r"""Compose a command for running a program using the exectuable for
        this language (compiler/interpreter) with the provided arguments.

        Args:
            args (list): The program that returned command should run and any
                arguments that should be provided to it.
            unused_kwargs (dict, optional): Existing dictionary that unused
                keyword arguments should be added to. Defaults to None and is
                ignored.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Arguments composing the command required to run the program
                from the command line using the executable for this language.

        """
        if isinstance(unused_kwargs, dict):
            unused_kwargs.update(kwargs)
        return args
        
    @classmethod
    def is_language_installed(cls):
        r"""Determine if the interpreter/compiler for the associated programming
        language is installed.

        Returns:
            bool: True if the language interpreter/compiler is installed.

        """
        return True

    @classmethod
    def is_configured(cls):
        r"""Determine if the appropriate configuration has been performed (e.g.
        installation of supporting libraries etc.)

        Returns:
            bool: True if the language has been configured.

        """
        return True

    @classmethod
    def is_comm_installed(self, commtype=None, **kwargs):
        r"""Determine if a comm is installed for the associated programming
        language.

        Returns:
            bool: True if a comm is installed for this language.

        """
        if commtype is None:
            return True  # executables presumed to include comms
        return (commtype in tools.get_supported_comm())



import yggdrasil.drivers.tests.test_InterpretedModelDriver as parent


class TestLPyModelParam(parent.TestInterpretedModelParam):
    r"""Test parameters for LPyModelDriver class."""

    driver = 'LPyModelDriver'
    
    def tests_on_not_installed(self):
        r"""Tests for when the driver is not installed."""
        super(TestLPyModelParam, self).tests_on_not_installed()
        self.assert_raises(RuntimeError, self.import_cls.language_version)

        
class TestLPyModelDriverNoInit(TestLPyModelParam,
                               parent.TestInterpretedModelDriverNoInit):
    r"""Test runner for LPyModelDriver class without initing the driver."""
    pass


class TestLPyModelDriverNoStart(TestLPyModelParam,
                                parent.TestInterpretedModelDriverNoStart):
    r"""Test runner for LPyModelDriver class without starting the driver."""
    pass


class TestLPyModelDriver(TestLPyModelParam,
                         parent.TestInterpretedModelDriver):
    r"""Test runner for LPyModelDriver class."""
    pass



import unittest
from yggdrasil.communication.RMQComm import RMQComm
import yggdrasil.drivers.tests.test_ServerDriver as parent


_rmq_installed = RMQComm.is_installed(language='python')


@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQAsyncServerParam(parent.TestServerParam):
    r"""Test parameters for RMQAsyncServerDriver class."""

    def __init__(self, *args, **kwargs):
        super(TestRMQAsyncServerParam, self).__init__(*args, **kwargs)
        self.driver = 'RMQAsyncServerDriver'
        self.client_comm = 'RMQAsyncComm'
        self.icomm_name = self.client_comm
        self.timeout = 10.0
        self.route_timeout = 60.0

    
@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQAsyncServerDriverNoStart(TestRMQAsyncServerParam,
                                      parent.TestServerDriverNoStart):
    r"""Test class for RMQAsyncServerDriver class without start."""
    pass


@unittest.skipIf(not _rmq_installed, "RMQ Server not running")
class TestRMQAsyncServerDriver(TestRMQAsyncServerParam, parent.TestServerDriver):
    r"""Test class for RMQAsyncServerDriver class."""
    pass



from yggdrasil.metaschema.properties.MetaschemaProperty import (
    MetaschemaProperty)


class LengthMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'length' property."""
    
    name = 'length'
    schema = {'description': 'Number of elements in the 1D array.',
              'type': 'number',
              'minimum': 1}

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'length' 1darray property."""
        return len(instance)


class ShapeMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'shape' property."""
    
    name = 'shape'
    schema = {'description': 'Shape of the ND array in each dimension.',
              'type': 'array',
              'items': {'type': 'integer',
                        'minimum': 1}}

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'shape' ndarray property."""
        return list(instance.shape)

    @classmethod
    def compare(cls, prop1, prop2, root1=None, root2=None):
        r"""Comparison method for the 'shape' ndarray property."""
        if len(prop1) != len(prop2):
            yield '%d dimensions not compatible with %d dimensions' % (
                len(prop1), len(prop2))
        else:
            for i, (p1, p2) in enumerate(zip(prop1, prop2)):
                if p1 != p2:
                    yield "Size in dimension %d dosn't match. %d vs. %s" % (i, p1, p2)



import numpy as np
import pandas as pd
from collections import OrderedDict
import yggdrasil.drivers.tests.test_InterpretedModelDriver as parent


class TestRModelParam(parent.TestInterpretedModelParam):
    r"""Test parameters for RModelDriver."""

    driver = "RModelDriver"


class TestRModelDriverNoInit(TestRModelParam,
                             parent.TestInterpretedModelDriverNoInit):
    r"""Test runner for RModelDriver without init."""

    def test_is_library_installed(self):
        r"""Test is_library_installed for invalid library."""
        self.assert_equal(
            self.import_cls.is_library_installed('invalid_unicorn'),
            False)

    def test_language2python(self):
        r"""Test language2python."""
        test_vars = [('hello', b'hello'),
                     (b'hello', b'hello'),
                     (u'hello', b'hello'),
                     (('hello', ), (b'hello', )),
                     (['hello'], [b'hello']),
                     ({'a': 'hello'}, {'a': b'hello'})]
        for a, b in test_vars:
            self.assert_equal(self.import_cls.language2python(a), b)

    def test_python2language(self):
        r"""Test python2language."""
        test_vars = [('hello', 'hello'),
                     (b'hello', 'hello'),
                     (u'hello', 'hello'),
                     (np.string_('hello'), 'hello'),
                     ((b'hello', ), ('hello', )),
                     ([b'hello'], ['hello']),
                     ({b'a': b'hello'}, {'a': 'hello'}),
                     (OrderedDict([(b'a', b'hello')]),
                      OrderedDict([('a', 'hello')]))]
        test_vars.append((
            pd.DataFrame.from_dict({'a': np.zeros(5, dtype='int64'),
                                    'b': np.ones(5, dtype=bytes)}),
            pd.DataFrame.from_dict({'a': np.zeros(5, dtype='int32'),
                                    'b': np.ones(5, dtype=str)})))
        for a, b in test_vars:
            self.assert_equal(self.import_cls.python2language(a), b)


class TestRModelDriverNoStart(TestRModelParam,
                              parent.TestInterpretedModelDriverNoStart):
    r"""Test runner for RModelDriver without start."""
    pass


class TestRModelDriver(TestRModelParam,
                       parent.TestInterpretedModelDriver):
    r"""Test runner for RModelDriver."""
    pass



from __future__ import print_function
# Import necessary connection interfaces
from yggdrasil.tools import print_encoded
from yggdrasil.interface.YggInterface import (
    YggAsciiFileInput, YggAsciiFileOutput,
    YggAsciiTableInput, YggAsciiTableOutput)

    
if __name__ == '__main__':

    # Input & output to an ASCII file line by line
    in_file = YggAsciiFileInput('inputPy_file')
    out_file = YggAsciiFileOutput('outputPy_file')
    # Input & output from a table row by row
    in_table = YggAsciiTableInput('inputPy_table')
    out_table = YggAsciiTableOutput('outputPy_table',
                                    '%5s\t%ld\t%3.1f\t%3.1lf%+3.1lfj\n')
    # Input & output from a table as an array
    in_array = YggAsciiTableInput('inputPy_array', as_array=True)
    out_array = YggAsciiTableOutput('outputPy_array',
                                    '%5s\t%ld\t%3.1f\t%3.1lf%+3.1lfj\n',
                                    as_array=True)

    # Read lines from ASCII text file until end of file is reached.
    # As each line is received, it is then sent to the output ASCII file.
    print('ascii_io(P): Receiving/sending ASCII file.')
    ret = True
    while ret:
        # Receive a single line
        (ret, line) = in_file.recv()
        if ret:
            # If the receive was succesful, send the line to output
            print("File: %s" % line)
            ret = out_file.send(line)
            if not ret:
                raise RuntimeError("ascii_io(P): ERROR SENDING LINE")
        else:
            # If the receive was not succesful, send the end-of-file message to
            # close the output file.
            print("End of file input (Python)")
            out_file.send_eof()

    # Read rows from ASCII table until end of file is reached.
    # As each row is received, it is then sent to the output ASCII table
    print('ascii_io(P): Receiving/sending ASCII table.')
    ret = True
    while ret:
        # Receive a single row
        (ret, line) = in_table.recv()
        if ret:
            # If the receive was succesful, send the values to output.
            # Formatting is taken care of on the output driver side.
            print_encoded("Table: %s, %d, %3.1f, %s" % tuple(line))
            # print("Table: %s, %d, %3.1f, %s" % line)
            ret = out_table.send(*line)
            if not ret:
                raise RuntimeError("ascii_io(P): ERROR SENDING ROW")
        else:
            # If the receive was not succesful, send the end-of-file message to
            # close the output file.
            print("End of table input (Python)")
            out_table.send_eof()

    # Read entire array from ASCII table into numpy array
    ret = True
    while ret:
        ret, arr = in_array.recv_array()
        if ret:
            print("Array: (%d rows)" % len(arr))
            # Print each line in the array
            for i in range(len(arr)):
                print("%5s, %d, %3.1f, %s" % tuple(arr[i]))
            # Send the array to output. Formatting is handled on the output driver side.
            ret = out_array.send_array(arr)
            if not ret:
                raise RuntimeError("ascii_io(P): ERROR SENDING ARRAY")
        else:
            print("ascii_io(P): End of array input (Python)")



import sys
import logging
from subprocess import Popen, PIPE
from yggdrasil import platform, tools
from yggdrasil.communication import CommBase, AsyncComm
logger = logging.getLogger(__name__)
try:
    import sysv_ipc
    _ipc_installed = (platform._is_linux or platform._is_mac)
except ImportError:  # pragma: windows
    logger.warn("Could not import sysv_ipc. "
                + "IPC support will be disabled.")
    sysv_ipc = None
    _ipc_installed = False


def get_queue(qid=None):
    r"""Create or return a sysv_ipc.MessageQueue and register it.

    Args:
        qid (int, optional): If provided, ID for existing queue that should be
           returned. Defaults to None and a new queue is returned.

    Returns:
        :class:`sysv_ipc.MessageQueue`: Message queue.

    """
    if _ipc_installed:
        kwargs = dict(max_message_size=tools.get_YGG_MSG_MAX())
        if qid is None:
            kwargs['flags'] = sysv_ipc.IPC_CREX
        mq = sysv_ipc.MessageQueue(qid, **kwargs)
        key = str(mq.key)
        CommBase.register_comm('IPCComm', key, mq)
        return mq
    else:  # pragma: windows
        logger.warning("IPC not installed. Queue cannot be returned.")
        return None


def remove_queue(mq):
    r"""Remove a sysv_ipc.MessageQueue and unregister it.

    Args:
        mq (:class:`sysv_ipc.MessageQueue`) Message queue.
    
    Raises:
        KeyError: If the provided queue is not registered.

    """
    key = str(mq.key)
    if not CommBase.is_registered('IPCComm', key):
        raise KeyError("Queue not registered.")
    CommBase.unregister_comm('IPCComm', key)
    

def ipcs(options=[]):
    r"""Get the output from running the ipcs command.

    Args:
        options (list): List of flags that should be used. Defaults to an empty
            list.

    Returns:
        str: Captured output.

    """
    if _ipc_installed:
        cmd = ' '.join(['ipcs'] + options)
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
        output, err = p.communicate()
        exit_code = p.returncode
        if exit_code != 0:  # pragma: debug
            if not err.isspace():
                print(err.decode('utf-8'))
            raise Exception("Error on spawned process. See output.")
        return output.decode('utf-8')
    else:  # pragma: windows
        logger.warn("IPC not installed. ipcs cannot be run.")
        return ''


def ipc_queues():
    r"""Get a list of active IPC queues.

    Returns:
       list: List of IPC queues.

    """
    skip_lines = [
        # Linux
        '------ Message Queues --------',
        'key        msqid      owner      perms      used-bytes   messages    ',
        # MacOS
        'IPC status from',
        'Message Queues:',
        'T     ID     KEY        MODE       OWNER    GROUP']
    out = ipcs(['-q']).split('\n')
    qlist = []
    for l in out:
        skip = False
        if len(l) == 0:
            skip = True
        else:
            for ls in skip_lines:
                if ls in l:
                    skip = True
                    break
        if not skip:
            if platform._is_linux:
                key_col = 0
            elif platform._is_mac:
                key_col = 2
            else:  # pragma: debug
                raise NotImplementedError("Unsure what column the queue key "
                                          + "is in on this platform "
                                          + "(%s)" % sys.platform)
            qlist.append(l.split()[key_col])
    return qlist


def ipcrm(options=[]):
    r"""Remove IPC constructs using the ipcrm command.

    Args:
        options (list): List of flags that should be used. Defaults to an empty
            list.

    """
    if _ipc_installed:
        cmd = ' '.join(['ipcrm'] + options)
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=True)
        output, err = p.communicate()
        exit_code = p.returncode
        if exit_code != 0:  # pragma: debug
            if not err.isspace():
                print(err.decode('utf-8'))
            raise Exception("Error on spawned process. See output.")
        if not output.isspace():
            print(output.decode('utf-8'))
    else:  # pragma: windows
        logger.warn("IPC not installed. ipcrm cannot be run.")


def ipcrm_queues(queue_keys=None):
    r"""Delete existing IPC queues.

    Args:
        queue_keys (list, str, optional): A list of keys for queues that should
            be removed. Defaults to all existing queues.

    """
    if _ipc_installed:
        if queue_keys is None:
            queue_keys = ipc_queues()
        if isinstance(queue_keys, str):
            queue_keys = [queue_keys]
        for q in queue_keys:
            ipcrm(["-Q %s" % q])
    else:  # pragma: windows
        logger.warn("IPC not installed. ipcrm cannot be run.")


class IPCServer(CommBase.CommServer):
    r"""IPC server object for cleaning up server queue."""

    def terminate(self, *args, **kwargs):
        CommBase.unregister_comm('IPCComm', self.srv_address)
        super(IPCServer, self).terminate(*args, **kwargs)


class IPCComm(AsyncComm.AsyncComm):
    r"""Class for handling I/O via IPC message queues.

    Attributes:
        q (:class:`sysv_ipc.MessageQueue`): Message queue.

    Developer Notes:
        The default size limit for IPC message queues is 2048 bytes on Mac
        operating systems so it is important that implementation of this
        communication mechanism properly split and send messages larger than
        this limit as more than one message.
        
    """

    _commtype = 'ipc'
    _schema_subtype_description = ('Interprocess communication (IPC) queue.')
    _maxMsgSize = 2048  # Based on IPC limit on MacOS
    address_description = ("An IPC message queue key.")

    def _init_before_open(self, **kwargs):
        r"""Initialize empty queue and server class."""
        self.q = None
        self._server_class = IPCServer
        super(IPCComm, self)._init_before_open(**kwargs)
            
    @classmethod
    def underlying_comm_class(self):
        r"""str: Name of underlying communication class."""
        return 'IPCComm'

    @classmethod
    def close_registry_entry(cls, value):
        r"""Close a registry entry."""
        try:
            value.remove()
            out = True
        except sysv_ipc.ExistentialError:  # pragma: debug
            out = False
        return out

    @classmethod
    def new_comm_kwargs(cls, *args, **kwargs):
        r"""Initialize communication with new queue."""
        if 'address' not in kwargs:
            kwargs.setdefault('address', 'generate')
        return args, kwargs

    def bind(self):
        r"""Bind to random queue if address is generate."""
        if not self._bound:
            if self.address == 'generate':
                self._bound = True
                q = get_queue()
                self.address = str(q.key)
        super(IPCComm, self).bind()

    def open_after_bind(self):
        r"""Open the connection by getting the queue from the bound address."""
        qid = int(self.address)
        self.q = get_queue(qid)

    def _open_direct(self):
        r"""Open the queue."""
        if not self.is_open_direct:
            self.bind()
            self.open_after_bind()
            self.debug("qid: %s", self.q.key)

    def _close_direct(self, skip_remove=False):
        r"""Close the queue."""
        if self._bound and (self.q is None):
            try:
                self.open_after_bind()
            except sysv_ipc.ExistentialError:  # pragma: debug
                self.q = None
                self._bound = False
        # Remove the queue
        dont_close = (skip_remove or self.is_client)
        if (self.q is not None) and (not dont_close):
            # Dont close for client because server will not be able
            # to unregister the comm
            self.unregister_comm(self.address, dont_close=dont_close)
        self.q = None
        self._bound = False

    @property
    def is_open_direct(self):
        r"""bool: True if the queue is not None."""
        if self.q is None:
            return False
        try:
            self.q.current_messages
        except AttributeError:  # pragma: debug
            if self.q is not None:
                raise
            return False
        except sysv_ipc.ExistentialError:  # pragma: debug
            self._close_direct()
            return False
        return True

    def confirm_send(self, noblock=False):
        r"""Confirm that sent message was received."""
        if noblock:
            return True
        return (self.n_msg_direct_send == 0)

    def confirm_recv(self, noblock=False):
        r"""Confirm that message was received."""
        return True

    @property
    def n_msg_direct_send(self):
        r"""int: Number of messages in the queue to send."""
        if self.is_open_direct:
            try:
                return self.q.current_messages
            except AttributeError:  # pragma: debug
                if self.is_open_direct:
                    raise
                return 0
            except sysv_ipc.ExistentialError:  # pragma: debug
                self._close_direct()
                return 0
        else:
            return 0
        
    @property
    def n_msg_direct_recv(self):
        r"""int: Number of messages in the queue to recv."""
        return self.n_msg_direct_send

    def _send_direct(self, payload):
        r"""Send a message to the comm directly.

        Args:
            payload (str): Message to send.

        Returns:
            bool: Success or failure of sending the message.

        """
        if not self.is_open_direct:  # pragma: debug
            return False
        try:
            self.debug('Sending %d bytes', len(payload))
            self.q.send(payload, block=False)
            self.debug('Sent %d bytes', len(payload))
        except sysv_ipc.BusyError:  # pragma: debug
            self.debug("IPC Queue Full")
            raise AsyncComm.AsyncTryAgain
        except OSError:  # pragma: debug
            self.debug("Send failed")
            self._close_direct()
            return False
        except AttributeError:  # pragma: debug
            if self.is_closed:
                self.debug("Comm closed")
                return False
            raise
        return True

    def _recv_direct(self):
        r"""Receive a message from the comm directly.

        Returns:
            tuple (bool, str): The success or failure of receiving a message
                and the message received.

        """
        # Receive message
        self.debug("Message ready, reading it.")
        try:
            data, _ = self.q.receive()  # ignore ident
            self.debug("Received %d bytes", len(data))
        except sysv_ipc.ExistentialError:  # pragma: debug
            self.debug("sysv_ipc.ExistentialError: closing")
            self._close_direct()
            return (False, self.empty_bytes_msg)
        except AttributeError:  # pragma: debug
            if self.is_closed:
                self.debug("Queue closed")
                return (False, self.empty_bytes_msg)
            raise
        return (True, data)

    def purge(self):
        r"""Purge all messages from the comm."""
        super(IPCComm, self).purge()
        try:
            while self.n_msg_direct > 0:  # pragma: debug
                self.q.receive()
        except AttributeError:  # pragma: debug
            if self.is_open_direct:
                raise



from yggdrasil.communication import FileComm


class AsciiMapComm(FileComm.FileComm):
    r"""Class for handling I/O from/to a ASCII map on disk.

    Args:
        name (str): The environment variable where file path is stored.
        **kwargs: Additional keywords arguments are passed to parent class.

    """

    _filetype = 'map'
    _schema_subtype_description = ('The file contains a key/value mapping '
                                   'with one key/value pair per line and '
                                   'separated by some delimiter.')
    _default_serializer = 'map'



from yggdrasil.communication.tests import test_FileComm as parent


class TestPickleFileComm(parent.TestFileComm):
    r"""Test for PickleFileComm communication class."""

    comm = 'PickleFileComm'



from yggdrasil.serialize.tests import test_SerializeBase as parent


class TestPickleSerialize(parent.TestSerializeBase):
    r"""Test class for TestPickleSerialize class."""

    _cls = 'PickleSerialize'

    def test_get_first_frame(self):
        r"""Test get_first_frame for empty message."""
        self.assert_equal(self.import_cls.get_first_frame(b'not a pickle'), b'')



import numpy as np
from yggdrasil.serialize.tests import test_SerializeBase as parent


class TestDefaultSerialize(parent.TestSerializeBase):
    r"""Test class for DefaultSerialize class."""

    _cls = 'DefaultSerialize'
        
    def test_serialize_no_format(self):
        r"""Test serialize/deserialize without format string."""
        if (len(self._inst_kwargs) == 0) and (self._cls == 'DefaultSerialize'):
            for iobj in self.testing_options['objects']:
                msg = self.instance.serialize(iobj,
                                              header_kwargs=self._header_info)
                iout, ihead = self.instance.deserialize(msg)
                self.assert_result_equal(iout, iobj)
                # self.assert_equal(ihead, self._header_info)
            # self.assert_raises(Exception, self.instance.serialize, ['msg', 0])

    def test_invalid_update(self):
        r"""Test error raised when serializer updated with type that isn't
        compatible."""
        if (len(self._inst_kwargs) == 0) and (self._cls == 'DefaultSerialize'):
            self.instance.initialize_from_message(np.int64(1))
            self.assert_raises(RuntimeError, self.instance.update_serializer,
                               type='ply')
        

class TestDefaultSerialize_format(TestDefaultSerialize):
    r"""Test class for DefaultSerialize class with format."""

    testing_option_kws = {'table_example': True, 'include_oldkws': True}


class TestDefaultSerialize_array(TestDefaultSerialize_format):
    r"""Test class for DefaultSerialize class with format as array."""

    testing_option_kws = {'table_example': True, 'include_oldkws': True,
                          'array_columns': True}


class TestDefaultSerialize_uniform(TestDefaultSerialize):
    r"""Test class for items as dictionary."""
    
    def get_options(self):
        r"""Get testing options."""
        out = {'kwargs': {'type': 'array', 'items': {'type': '1darray',
                                                     'subtype': 'float',
                                                     'precision': 64}},
               'empty': [],
               'objects': [[np.zeros(3, 'float'), np.zeros(3, 'float')],
                           [np.ones(3, 'float'), np.ones(3, 'float')]],
               'extra_kwargs': {},
               'typedef': {'type': 'array', 'items': {'type': '1darray',
                                                      'subtype': 'float',
                                                      'precision': 64}},
               'dtype': np.dtype("float64")}
        return out


class TestDefaultSerialize_uniform_names(TestDefaultSerialize_uniform):
    r"""Test class for items as dictionary."""
    
    def get_options(self):
        r"""Get testing options."""
        out = super(TestDefaultSerialize_uniform_names, self).get_options()
        out['kwargs']['field_names'] = [b'a', b'b']
        out['kwargs']['field_units'] = [b'cm', b'g']
        out['field_names'] = ['a', 'b']
        out['field_units'] = ['cm', 'g']
        out['dtype'] = np.dtype([('a', '<f8'), ('b', '<f8')])
        out['typedef'] = {'type': 'array',
                          'items': [{'type': '1darray',
                                     'subtype': 'float',
                                     'precision': 64,
                                     'title': 'a',
                                     'units': 'cm'},
                                    {'type': '1darray',
                                     'subtype': 'float',
                                     'precision': 64,
                                     'title': 'b',
                                     'units': 'g'}]}
        return out
    

class TestDefaultSerialize_type(TestDefaultSerialize):
    r"""Test class for DefaultSerialize class with types."""

    def get_options(self):
        r"""Get testing options."""
        out = {'kwargs': {'type': 'float'},
               'empty': b'',
               'objects': [float(x) for x in range(5)],
               'extra_kwargs': {},
               'typedef': {'type': 'float'},
               'dtype': None}
        return out



"""
Small scanf implementation.

Python has powerful regular expressions but sometimes they are totally overkill
when you just want to parse a simple-formatted string.
C programmers use the scanf-function for these tasks (see link below).

This implementation of scanf translates the simple scanf-format into
regular expressions. Unlike C you can be sure that there are no buffer overflows
possible.


For more information see
  * http://www.python.org/doc/current/lib/node49.html
  * http://en.wikipedia.org/wiki/Scanf

Original code from:
    http://code.activestate.com/recipes/502213-simple-scanf-implementation/

Modified original to make the %f more robust, as well as added %* modifier to
skip fields.

Version: 1.3.3

Releases:
  1.0
    2010-10-11
      * Initial release

  1.1
    2010-10-13
      * Changed regex from 'match' (only matches at beginning of line)
        to 'search' (matches anywhere in line)
      * Bugfix - ignore cast for skipped fields

  1.2
    2013-05-30
      * Added 'collapseWhitespace' flag (defaults to True) to take the search
        string and replace all whitespace with regex string to match repeated
        whitespace.  This enables better matching in log files where the data
        has been formatted for easier reading.  These cases have variable
        amounts of whitespace between the columns, depending on the number
        of characters in the data itself.
      
  1.3
    2016-01-18
      * Add 'extractdata' function.
    
  1.3.1
    2016-06-23
      * Release to PyPi, now including README.md

Updates by Meagan Lang:
  2018-04-12
    * Added ability to parse components of field specifiers and cope with
      whitespace padding.

"""
import re
import sys

__version__ = '1.3.3'

__all__ = ["scanf", 'scanf_translate', 'scanf_compile']


DEBUG = False


def no_cast(s):
    r"""Dummy function to cast a string to a string.

    Args:
        s (str): String to cast.

    Returns:
        str: Cast string.

    """
    return s


def cast_str(s):
    r"""Cast a string value to a string, stripping whitespace.

    Args:
        s (str): String to cast.

    Returns:
        str: Cast string.

    """
    return s.strip()


def cast_hex(s):
    r"""Cast a string value to a hex integer.

    Args:
        s (str): String to cast.

    Returns:
        int: Base 16 integer for provided string.

    """
    return int(s, 16)


def cast_oct(s):
    r"""Cast a string value to an octal.

    Args:
        s (str): String to cast.

    Returns:
        int: Base 8 integer for provided string.

    """
    return int(s, 8)


def cformat2regex(flags, width, precision, length, specifier):
    r"""Get regex substring that will match the given cformat components.

    Args:
        flags (str): Format flags.
        width (str): Minimum field width.
        precision (str): Field precision.
        length (str): Field value size (e.g. short, long).
        specifier (str): Field specifier.

    Returns:
        str: Regex expression that will match the provided components.

    """
    pat_dec = '[-+]?\\d+(?:\\.\\d+)?'
    pat_sub = ''
    pad_zero = 0
    # Left padding specified in flags
    if '0' in flags:
        if '+' in flags:
            pat_sub += '[-+]'
        if width and specifier in 'diueEfgGoxX':
            pad_zero = int(width)
    else:
        if ('-' not in flags) and width and (specifier != 's'):
            pat_sub += "\\s{,%s}" % width
        if '+' in flags:
            pat_sub += '[-+]'
    if precision and specifier in 'diuoxX':
        pad_zero = max(pad_zero, int(precision[1:]))
    # Casts
    if pad_zero and specifier in 'diueEfgG':
        pat_sub += '0{,%d}' % pad_zero
    if specifier == 'f':
        pat_sub += pat_dec
    elif specifier in 'eE':
        pat_sub += '%s%s[+-]\\d+' % (pat_dec, specifier)
    elif specifier in 'gG':
        pat_sub += '%s(?:[eE][+-]\\d+)?' % (pat_dec)
    elif specifier in 'diu':
        pat_sub += '\\d+'
    elif specifier in 'cs':
        if not width:
            if specifier == 'c':
                pat_sub += '.'
            else:
                pat_sub += '\\S+'
        else:
            pat_sub += '.{%s}' % width
    elif specifier in 'xX':
        if '#' in flags:
            pat_sub += '0%s' % specifier
        if pad_zero:
            pat_sub += '0{,%d}' % pad_zero
        pat_sub += '[\\dA-Za-f]+'
    elif specifier == 'o':
        if '#' in flags:
            pat_sub += '0'
        if pad_zero:
            pat_sub += '0{,%d}' % pad_zero
        pat_sub += '[0-7]*'
    if ('-' in flags) and ('0' not in flags) and width:
        pat_sub += "\\s{,%s}" % width
    return pat_sub


def parse_cformat(format, i):
    pattern = None
    cast = None
    # %[flags][width][.precision][length]specifier
    # float_format = "%([ -\\+0]{,4})(\\d*)((?:\\.\\d+)?)(L?)([eEfgG])"
    # First check for match to complex
    complex_format = ("%([ -0]{,3})(\\d*)((?:\\.\\d+)?)([hlL]{,2})([eEfgG])"
                      + "%(\\+?[ -0]{,3}\\+?)(\\d*)((?:\\.\\d+)?)([hlL]{,2})([eEfgG])j")
    token = re.compile(complex_format)
    found = token.match(format, i)
    if found:
        cast = complex
        groups = found.groups()
        pat_sub1 = cformat2regex(*groups[:5])
        pat_sub2 = cformat2regex(*groups[5:])
        pattern = "(%s%sj)" % (pat_sub1, pat_sub2)
        return found, pattern, cast
    # Then check for generic
    any_format = "%([\\* \\-\\+#0]{,6})(\\d*)((?:\\.\\d+)?)([hlL]{,2})([cdieEfgGosuxX])"
    token = re.compile(any_format)
    found = token.match(format, i)
    if found:
        groups = found.groupdict() or found.groups()
        # Get cast
        specifier = groups[-1]
        if specifier in 'eEfgG':
            cast = float
        elif specifier in 'diu':
            cast = int
        elif specifier == 'c':
            cast = no_cast
        elif specifier == 's':
            cast = cast_str
        elif specifier in ['x', 'X']:
            cast = cast_hex
        elif specifier in ['o']:
            cast = cast_oct
        else:  # pragma: debug
            raise Exception('No cast for specifier %s.' % specifier)
        # Get pattern
        pat_sub = cformat2regex(*groups)
        if '*' in groups[0]:
            pattern = "(?:%s)" % pat_sub
            cast = None
        else:
            pattern = "(%s)" % pat_sub
        return found, pattern, cast
    return found, pattern, cast


# As you can probably see it is relatively easy to add more format types.
# Make sure you add a second entry for each new item that adds the extra
#   few characters needed to handle the field ommision.
scanf_translate = [
    (re.compile(_token), _pattern, _cast) for _token, _pattern, _cast in [
        ("%c", "(.)", lambda x:x),
        ("%\\*c", "(?:.)", None),
        
        ("%(\\d)c", "(.{%s})", lambda x:x),
        ("%\\*(\\d)c", "(?:.{%s})", None),
        
        ("%(\\d)[di]", "([+-]?\\d{%s})", int),
        ("%\\*(\\d)[di]", "(?:[+-]?\\d{%s})", None),
        
        ("%-(\\d)[di]", "(.{%s})", int),
        ("%\\*-(\\d)[di]", "(?:.{%s})", None),
        
        ("%[di]", "([+-]?\\d+)", int),
        ("%\\*[di]", "(?:[+-]?\\d+)", None),
        
        ("%u", "(\\d+)", int),
        ("%\\*u", "(?:\\d+)", None),
        
        # langmm: complex
        ("%[fgeE]%[+-][fgeE]j",
         "("
         + "(?:[-+]?(?:\\d+(?:\\.\\d*)?|\\.\\d+)(?:[eE][-+]?\\d+)?)"
         + "[+-]"
         + "(?:(?:\\d+(?:\\.\\d*)?|\\.\\d+)(?:[eE][-+]?\\d+)?)"
         + "j)",
         complex),
        ("%\\*[fgeE]%[+-][fgeE]j",
         "(?:"
         + "(?:[-+]?(?:\\d+(?:\\.\\d*)?|\\.\\d+)(?:[eE][-+]?\\d+)?)"
         + "[+-]"
         + "(?:(?:\\d+(?:\\.\\d*)?|\\.\\d+)(?:[eE][-+]?\\d+)?)"
         + "j)",
         None),

        ("%[fgeE]", "([-+]?(?:\\d+(?:\\.\\d*)?|\\.\\d+)(?:[eE][-+]?\\d+)?)", float),
        ("%\\*[fgeE]", "(?:[-+]?(?:\\d+(?:\\.\\d*)?|\\.\\d+)(?:[eE][-+]?\\d+)?)", None),
        
        # langmm: Added to allows matching of %5s
        ("%(\\d)s", "(.{%s})", lambda x:x.strip()),
        ("%\\*(\\d)s", "(?:.{%s})", None),
        
        ("%s", "(\\S+)", lambda x:x),
        ("%\\*s", "(?:\\S+)", None),
        
        ("%([xX])", "(0%s[\\dA-Za-f]+)", lambda x:int(x, 16)),
        ("%\\*([xX])", "(?:0%s[\\dA-Za-f]+)", None),
        
        ("%o", "(0[0-7]*)", lambda x:int(x, 8)),
        ("%\\*o", "(?:0[0-7]*)", None),
    ]]


# Cache formats
SCANF_CACHE_SIZE = 1000
scanf_cache = {}


def scanf_compile(format, collapseWhitespace=True):
    """
    Translate the format into a regular expression

    For example:
    >>> format_re, casts = _scanf_compile('%s - %d errors, %d warnings')
    >>> print format_re.pattern
    (\\S+) \\- ([+-]?\\d+) errors, ([+-]?\\d+) warnings

    Translated formats are cached for faster reuse
    """
    compiled = scanf_cache.get(format)
    if compiled:
        return compiled

    format_pat = ""
    cast_list = []
    i = 0
    length = len(format)
    while i < length:
        found = None
        found, pattern, cast = parse_cformat(format, i)
        if found:
            if cast:
                cast_list.append(cast)
            format_pat += pattern
            i = found.end()
        # for token, pattern, cast in scanf_translate:
        #     found = token.match(format, i)
        #     if found:
        #         print('found', pattern)
        #         if cast:  # cast != None
        #             cast_list.append(cast)
        #         groups = found.groupdict() or found.groups()
        #         if groups:
        #             pattern = pattern % groups
        #         format_pat += pattern
        #         i = found.end()
        #         break
        if not found:
            char = format[i]
            # escape special characters
            if char in "()[]-.+*?{}<>\\":
                format_pat += "\\"
            format_pat += char
            i += 1
    if DEBUG:
        print("DEBUG: %r -> %s" % (format, format_pat))
    if collapseWhitespace:
        format_pat = re.sub('\\s+', r'\\s+', format_pat)

    format_re = re.compile(format_pat)
    if len(scanf_cache) > SCANF_CACHE_SIZE:
        scanf_cache.clear()
    scanf_cache[format] = (format_re, cast_list)
    return format_re, cast_list


def scanf(format, s=None, collapseWhitespace=True):
    """
    scanf supports the following formats:
      %c        One character
      %5c       5 characters
      %d, %i    int value
      %7d, %7i  int value with length 7
      %f        float value
      %o        octal value
      %X, %x    hex value
      %s        string terminated by whitespace

    Examples:
    >>> scanf("%s - %d errors, %d warnings", "/usr/sbin/sendmail - 0 errors, 4 warnings")
    ('/usr/sbin/sendmail', 0, 4)
    >>> scanf("%o %x %d", "0123 0x123 123")
    (66, 291, 123)


    If the parameter s is a file-like object, s.readline is called.
    If s is not specified, stdin is assumed.

    The function returns a tuple of found values
    or None if the format does not match.
    """

    if s is None:
        s = sys.stdin
    if hasattr(s, "readline"):
        s = s.readline()

    # print(s, format)

    format_re, casts = scanf_compile(format, collapseWhitespace)

    found = format_re.search(s)
    if found:
        groups = found.groups()
        return tuple([casts[i](groups[i]) for i in range(len(groups))])


def extractdata(pattern, text=None, filepath=None):
    """
    Read through an entire file or body of text one line at a time. Parse each
    line that matches the supplied pattern string and ignore the rest.
    
    If *text* is supplied, it will be parsed according to the *pattern* string.
    If *text* is not supplied, the file at *filepath* will be opened and parsed.
    """
    y = []
    if text is None:
        with open(filepath) as f:
            for line in f:
                match = scanf(pattern, line)
                if match:
                    if len(y) == 0:
                        y = [[float(s)] for s in match]
                    else:
                        for i, ydata in enumerate(y):
                            ydata.append(float(match[i]))
    else:
        for line in text.split('\n'):
            match = scanf(pattern, line)
            if match:
                if len(y) == 0:
                    y = [[float(s)] for s in match]
                else:
                    for i, ydata in enumerate(y):
                        ydata.append(float(match[i]))
                        
    return y


if __name__ == "__main__":
    import doctest
    doctest.testmod(verbose=True, report=True)



import copy
import numpy as np
from yggdrasil import serialize
from yggdrasil.tests import assert_equal
from yggdrasil.metaschema.datatypes.JSONArrayMetaschemaType import (
    JSONArrayMetaschemaType)
from yggdrasil.metaschema.datatypes.tests import test_MetaschemaType as parent
from yggdrasil.metaschema.datatypes.tests import (
    test_ContainerMetaschemaType as container_utils)


def test_coerce():
    r"""Test serialization of coerced types."""
    typedef = {'type': 'array', 'items': [{'type': '1darray',
                                           'subtype': 'float',
                                           'title': 'a',
                                           'precision': 64}]}
    x = JSONArrayMetaschemaType(**typedef)
    key_order = ['a']
    msg_recv = [np.zeros(3, 'float64')]
    msg_send_list = [msg_recv[0],
                     serialize.list2numpy(msg_recv, names=key_order),
                     serialize.list2pandas(msg_recv, names=key_order),
                     serialize.list2dict(msg_recv, names=key_order)]

    def do_send_recv(msg_send):
        msg_seri = x.serialize(msg_send, tyepdef=typedef, key_order=key_order)
        assert_equal(x.deserialize(msg_seri)[0], msg_recv)

    for y in msg_send_list:
        do_send_recv(y)
    

class TestJSONArrayMetaschemaType(parent.TestMetaschemaType):
    r"""Test class for JSONArrayMetaschemaType class with float."""

    _mod = 'JSONArrayMetaschemaType'
    _cls = 'JSONArrayMetaschemaType'

    def __init__(self, *args, **kwargs):
        super(TestJSONArrayMetaschemaType, self).__init__(*args, **kwargs)
        self._value = []
        self._fulldef = {'type': self.import_cls.name,
                         'items': []}
        self._typedef = {'items': []}
        for i in range(container_utils._count):
            self._value.append(container_utils._vallist[i])
            self._fulldef['items'].append(container_utils._deflist[i])
            self._typedef['items'].append(container_utils._typedef[i])
        self._valid_encoded = [self._fulldef]
        self._valid_decoded = [self._value]
        self._invalid_encoded += [{'type': self._fulldef['type'],
                                   'items': [self._fulldef['items'][0]]}]
        self._invalid_encoded.append(copy.deepcopy(self._fulldef))
        del self._invalid_encoded[-1]['items'][0]['type']
        self._invalid_encoded.append(copy.deepcopy(self._fulldef))
        self._invalid_encoded[-1]['items'][0]['type'] = 'invalid'
        self._compatible_objects = [(self._value, self._value, None)]
        self._valid_normalize += [('1, 1 ', ['1', '1'])]

    def test_encode_data_readable(self):
        r"""Test corner case of encode_data_readable."""
        self.import_cls.encode_data_readable(['1', '1'], {})

    def test_container_errors(self):
        r"""Test errors on container operations."""
        self.assert_raises(RuntimeError, self.import_cls._assign, [], 10, None)

    def test_item_dictionary(self):
        r"""Test dictionary as items value."""
        x = [1, 2, 3]
        typedef = {'type': 'array', 'items': {'type': 'int'}}
        self.import_cls.validate_instance(x, typedef)
        self.import_cls.encode_data(x, typedef)

    def test_validate_errors(self):
        r"""Test error on validation of non-structured array."""
        self.assert_raises(ValueError, self.import_cls.validate,
                           np.zeros(5), raise_errors=True)



import os
import copy
import tempfile
from yggdrasil import backwards, platform
from yggdrasil.serialize.SerializeBase import SerializeBase
from yggdrasil.communication import CommBase


class FileComm(CommBase.CommBase):
    r"""Class for handling I/O from/to a file on disk.

    >>> x = FileComm('test_send', address='test_file.txt', direction='send')
    >>> x.send('Test message')
    True
    >>> with open('test_file.txt', 'r') as fd:
    ...     print(fd.read())
    Test message
    >>> x = FileComm('test_recv', address='test_file.txt', direction='recv')
    >>> x.recv()
    (True, b'Test message')

    Args:
        name (str): The environment variable where communication address is
            stored.
        read_meth (str, optional): Method that should be used to read data
            from the file. Defaults to 'read'. Ignored if direction is 'send'.
        append (bool, optional): If True and writing, file is openned in append
            mode. If True and reading, file is kept open even if the end of the
            file is reached to allow for another process to write to the file in
            append mode. Defaults to False.
        in_temp (bool, optional): If True, the path will be considered relative
            to the platform temporary directory. Defaults to False.
        is_series (bool, optional): If True, input/output will be done to
            a series of files. If reading, each file will be processed until
            the end is reached. If writing, each output will be to a new
            file in the series. The addressed is assumed to contain a format
            for the index of the file. Defaults to False.
        wait_for_creation (float, optional): Time (in seconds) that should be
            waited before opening for the file to be created if it dosn't exist.
            Defaults to 0 s and file will attempt to be opened immediately.
        **kwargs: Additional keywords arguments are passed to parent class.

    Attributes:
        fd (file): File that should be read/written.
        read_meth (str): Method that should be used to read data from the file.
        append (bool): If True and writing, file is openned in append mode.
        in_temp (bool): If True, the path will be considered relative to the
            platform temporary directory.
        is_series (bool): If True, input/output will be done to a series of
            files. If reading, each file will be processed until the end is
            reached. If writing, each output will be to a new file in the series.
        platform_newline (str): String indicating a newline on the current
            platform.

    Raises:
        ValueError: If the read_meth is not one of the supported values.

    """

    _filetype = 'binary'
    _datatype = {'type': 'bytes'}
    _schema_type = 'file'
    _schema_subtype_key = 'filetype'
    _schema_subtype_description = ('The entire file is read/written all at '
                                   'once as bytes.')
    _schema_required = ['name', 'filetype', 'working_dir', 'serializer']
    _schema_properties = {
        'working_dir': {'type': 'string'},
        'filetype': {'type': 'string', 'default': _filetype,
                     'description': ('The type of file that will be read from '
                                     'or written to.')},
        'read_meth': {'type': 'string', 'default': 'read',
                      'enum': ['read', 'readline']},
        'append': {'type': 'boolean', 'default': False},
        'in_temp': {'type': 'boolean', 'default': False},
        'is_series': {'type': 'boolean', 'default': False},
        'wait_for_creation': {'type': 'float', 'default': 0.0},
        'serializer': {'oneOf': [{'$ref': '#/definitions/serializer'},
                                 {'type': 'instance',
                                  'class': SerializeBase}],
                       'default': {'seritype': 'direct'}}}
    _schema_excluded_from_inherit = ['commtype', 'datatype', 'read_meth',
                                     'serializer']
    _schema_excluded_from_class_validation = ['serializer']
    _default_serializer = 'direct'
    _default_extension = '.txt'
    is_file = True
    _maxMsgSize = 0

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('close_on_eof_send', True)
        kwargs['partner_language'] = None  # Files don't have partner comms
        return super(FileComm, self).__init__(*args, **kwargs)

    def _init_before_open(self, **kwargs):
        r"""Get absolute path and set attributes."""
        self.header_was_read = False
        self.header_was_written = False
        super(FileComm, self)._init_before_open(**kwargs)
        # Process file class keywords
        if not hasattr(self, '_fd'):
            self._fd = None
        self.platform_newline = platform._newline
        if self.in_temp:
            self.address = os.path.join(tempfile.gettempdir(), self.address)
        self.address = os.path.abspath(self.address)
        self._series_index = 0
        if self.append:
            self.disable_header()
        if 'read_meth' not in self._schema_properties:
            self.read_meth = self.serializer.read_meth
        assert(self.read_meth in ['read', 'readline'])
        # Force overwrite for concatenation in append mode
        if self.append:
            if self.direction == 'recv':
                self.close_on_eof_recv = False
            elif (not self.serializer.concats_as_str):
                self.append = 'ow'
        # Assert that keyword args match serialization parameters
        if not self.serializer.concats_as_str:
            assert(self.read_meth == 'read')
            assert(not self.serializer.is_framed)
            
    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration."""
        CommBase.CommBase.before_registration(cls)
        # Add serializer properties to schema
        if cls._filetype != 'binary':
            assert('serializer' not in cls._schema_properties)
            cls._schema_properties.update(
                cls._default_serializer_class._schema_properties)
            del cls._schema_properties['seritype']

    @classmethod
    def get_testing_options(cls, read_meth=None, **kwargs):
        r"""Method to return a dictionary of testing options for this class.

        Args:
            read_meth (str, optional): Read method that will be used by the test
                class. Defaults to None and will be set by the serialier.
            **kwargs: Additional keyword arguments are passed to the parent
                class's method and the serializers methods for determining the
                default read_meth and concatenating the sent objects into the
                objects that are expected to be received.

        Returns:
            dict: Dictionary of variables to use for testing. Key/value pairs:
                kwargs (dict): Keyword arguments for comms tested with the
                    provided content.
                send (list): List of objects to send to test file.
                recv (list): List of objects that will be received from a test
                    file that was sent the messages in 'send'.
                contents (bytes): Bytes contents of test file created by sending
                    the messages in 'send'.

        """
        out = super(FileComm, cls).get_testing_options(**kwargs)
        if 'read_meth' in cls._schema_properties:
            if read_meth is None:
                read_meth = cls._schema_properties['read_meth']['default']
            out['kwargs']['read_meth'] = read_meth
        if read_meth == 'readline':
            out['recv_partial'] = [[x] for x in out['recv']]
            if cls._default_serializer == 'direct':
                comment = backwards.as_bytes(
                    cls._schema_properties['comment']['default'] + 'Comment\n')
                out['send'].append(comment)
                out['contents'] += comment
                out['recv_partial'].append([])
        else:
            seri_cls = cls._default_serializer_class
            if seri_cls.concats_as_str:
                out['recv_partial'] = [[x] for x in out['recv']]
                out['recv'] = seri_cls.concatenate(out['recv'], **out['kwargs'])
            else:
                out['recv_partial'] = [[out['recv'][0]]]
                for i in range(1, len(out['recv'])):
                    out['recv_partial'].append(seri_cls.concatenate(
                        out['recv_partial'][-1] + [out['recv'][i]], **out['kwargs']))
                out['recv'] = copy.deepcopy(out['recv_partial'][-1])
        return out
        
    @classmethod
    def is_installed(cls, language=None):
        r"""Determine if the necessary libraries are installed for this
        communication class.

        Args:
            language (str, optional): Specific language that should be checked
                for compatibility. Defaults to None and all languages supported
                on the current platform will be checked.

        Returns:
            bool: Is the comm installed.

        """
        # Filesystem is implied
        return True

    @classmethod
    def underlying_comm_class(self):
        r"""str: Name of underlying communication class."""
        return 'FileComm'

    @classmethod
    def close_registry_entry(cls, value):
        r"""Close a registry entry."""
        out = False
        if not value.closed:  # pragma: debug
            value.close()
            out = True
        return out

    @classmethod
    def new_comm_kwargs(cls, *args, **kwargs):
        r"""Initialize communication with new queue."""
        kwargs.setdefault('address', 'file.txt')
        return args, kwargs

    @property
    def open_mode(self):
        r"""str: Mode that should be used to open the file."""
        if self.direction == 'recv':
            io_mode = 'r'
        elif self.append == 'ow':
            io_mode = 'r+'
        elif self.append:
            io_mode = 'a'
        else:
            io_mode = 'w'
        io_mode += 'b'
        return io_mode

    def opp_comm_kwargs(self):
        r"""Get keyword arguments to initialize communication with opposite
        comm object.

        Returns:
            dict: Keyword arguments for opposite comm object.

        """
        kwargs = super(FileComm, self).opp_comm_kwargs()
        kwargs['is_series'] = self.is_series
        return kwargs

    @property
    def registry_key(self):
        r"""str: String used to register the socket."""
        # return self.address
        return '%s_%s_%s' % (self.address, self.direction, self.uuid)

    # Methods related to header
    def enable_header(self):
        r"""Turn on header so that it will be written."""
        self.header_was_read = False
        self.header_was_written = False
        if self.serializer.has_header:
            self.serializer.enable_file_header()
    
    def disable_header(self):
        r"""Turn off header so that it will not be written."""
        self.header_was_read = True
        self.header_was_written = True
        if self.serializer.has_header:
            self.serializer.disable_file_header()
    
    def read_header(self):
        r"""Read header lines from the file and update serializer info."""
        if self.header_was_read:
            return
        if self.serializer.has_header:
            pos = self.record_position()
            self.serializer.deserialize_file_header(self.fd)
            self.change_position(*pos)
        self.header_was_read = True

    def write_header(self):
        r"""Write header lines to the file based on the serializer info."""
        if self.header_was_written:
            return
        if self.serializer.has_header:
            header_msg = self.serializer.serialize_file_header()
            if header_msg:
                self.fd.write(header_msg)
        self.header_was_written = True

    # Methods related to position in the file/series
    def record_position(self):
        r"""Record the current position in the file/series."""
        _rec_pos = self.fd.tell()
        _rec_ind = self._series_index
        return _rec_pos, _rec_ind, self.header_was_read, self.header_was_written

    def reset_position(self, truncate=False):
        r"""Move to the front of the file and allow header to be read again.

        Args:
            truncate (bool, optional): If True, the file will be truncated after
                moving to the beginning, effectively erasing the file. Defaults
                to False.

        """
        self.change_position(0)
        self.enable_header()
        if truncate:
            self.fd.truncate()

    def change_position(self, file_pos, series_index=None,
                        header_was_read=None, header_was_written=None):
        r"""Change the position in the file/series.

        Args:
            file_pos (int): Position that should be moved to in the file.
            series_index (int, optinal): Index of the file in the series that
                should be moved to. Defaults to None and will be set to the
                current series index.
            header_was_read (bool, optional): Status of if header has been
                read or not. Defaults to None and will be set to the current
                value.
            header_was_written (bool, optional): Status of if header has been
                written or not. Defaults to None and will be set to the current
                value.

        """
        if series_index is None:
            series_index = self._series_index
        if header_was_read is None:
            header_was_read = self.header_was_read
        if header_was_written is None:
            header_was_written = self.header_was_written
        self.advance_in_series(series_index)
        self.advance_in_file(file_pos)
        self.header_was_read = header_was_read
        self.header_was_written = header_was_written

    def advance_in_file(self, file_pos):
        r"""Advance to a certain position in the current file.

        Args:
            file_pos (int): Position that should be moved to in the current.
                file.

        """
        if self.is_open:
            try:
                self.fd.seek(file_pos)
            except (AttributeError, ValueError):  # pragma: debug
                if self.is_open:
                    raise

    def advance_in_series(self, series_index=None):
        r"""Advance to a certain file in a series.

        Args:
            series_index (int, optional): Index of file in the series that
                should be moved to. Defaults to None and call will advance to
                the next file in the series.

        Returns:
            bool: True if the file was advanced in the series, False otherwise.

        """
        out = False
        if self.is_series:
            if series_index is None:
                series_index = self._series_index + 1
            if self._series_index != series_index:
                if (((self.direction == 'send')
                     or os.path.isfile(self.get_series_address(series_index)))):
                    self._file_close()
                    self._series_index = series_index
                    self._open()
                    out = True
                    self.debug("Advanced to %d", series_index)
        if out:
            self.header_was_read = False
            self.header_was_written = False
        return out

    def get_series_address(self, index=None):
        r"""Get the address of a file in the series.

        Args:
            index (int, optional): Index in series to get address for.
                Defaults to None and the current index is used.

        Returns:
            str: Address for the file in the series.

        """
        if index is None:
            index = self._series_index
        return self.address % index

    @property
    def current_address(self):
        r"""str: Address of file currently being used."""
        if self.is_series:
            address = self.get_series_address()
        else:
            address = self.address
        return address

    # Methods related to opening/closing the file
    def _open(self):
        address = self.current_address
        if self.fd is None:
            if (not os.path.isfile(address)) and (self.wait_for_creation > 0):
                T = self.start_timeout(self.wait_for_creation)
                while (not T.is_out) and (not os.path.isfile(address)):
                    self.sleep()
                self.stop_timeout()
            self._fd = open(address, self.open_mode)
        T = self.start_timeout()
        while (not T.is_out) and (not self.is_open):  # pragma: debug
            self.sleep()
        self.stop_timeout()
        if self.append == 'ow':
            try:
                self.fd.seek(0, os.SEEK_END)
            except (AttributeError, ValueError):  # pragma: debug
                if self.is_open:
                    raise

    def _file_close(self):
        if self.is_open:
            try:
                self.fd.flush()
                os.fsync(self.fd.fileno())
            except OSError:  # pragma: debug
                pass
            try:
                self.fd.close()
            except (AttributeError, ValueError):  # pragma: debug
                if self.is_open:
                    raise
        self._fd = None

    def open(self):
        r"""Open the file."""
        super(FileComm, self).open()
        self._open()
        self.register_comm(self.registry_key, self.fd)

    def _close(self, *args, **kwargs):
        r"""Close the file."""
        self._file_close()
        self.unregister_comm(self.registry_key)
        super(FileComm, self)._close(*args, **kwargs)

    def remove_file(self):
        r"""Remove the file."""
        assert(self.is_closed)
        if self.is_series:
            i = 0
            while True:
                address = self.get_series_address(i)
                if not os.path.isfile(address):
                    break
                os.remove(address)
                i += 1
        else:
            if os.path.isfile(self.address):
                os.remove(self.address)

    @property
    def is_open(self):
        r"""bool: True if the connection is open."""
        try:
            return (self.fd is not None) and (not self.fd.closed)
        except AttributeError:  # pragma: debug
            if self.fd is not None:
                raise
            return False

    @property
    def fd(self):
        r"""Associated file identifier."""
        return self._fd

    @property
    def remaining_bytes(self):
        r"""int: Remaining bytes in the file."""
        if self.is_closed or self.direction == 'send':
            return 0
        pos = self.record_position()
        try:
            curpos = self.fd.tell()
            self.fd.seek(0, os.SEEK_END)
            endpos = self.fd.tell()
            out = endpos - curpos
        except (ValueError, AttributeError, OSError):  # pragma: debug
            if self.is_open:
                raise
            out = 0
        if self.is_series:
            i = self._series_index + 1
            while True:
                fname = self.get_series_address(i)
                if not os.path.isfile(fname):
                    break
                out += os.path.getsize(fname)
                i += 1
        self.change_position(*pos)
        return out

    @property
    def n_msg_recv(self):
        r"""int: The number of messages in the file."""
        if self.is_closed:
            return 0
        if self.read_meth == 'read':
            return int(self.remaining_bytes > 0)
        elif self.read_meth == 'readline':
            pos = self.record_position()
            try:
                out = 0
                flag, msg = self._recv()
                while len(msg) != 0 and msg != self.eof_msg:
                    out += 1
                    flag, msg = self._recv()
            except ValueError:  # pragma: debug
                out = 0
            self.change_position(*pos)
        else:  # pragma: debug
            self.error('Unsupported read_meth: %s', self.read_meth)
            out = 0
        return out

    def on_send_eof(self):
        r"""Close file when EOF to be sent.

        Returns:
            bool: False so that message not sent.

        """
        flag, msg_s = super(FileComm, self).on_send_eof()
        try:
            self.fd.flush()
        except (AttributeError, ValueError):  # pragma: debug
            if self.is_open:
                raise
        # self.close()
        return flag, msg_s

    def serialize(self, obj, **kwargs):
        r"""Serialize a message using the associated serializer."""
        if (not self.serializer.concats_as_str) and (self.fd.tell() != 0):
            new_obj = obj
            with open(self.current_address, 'rb') as fd:
                old_obj = self.deserialize(fd.read())[0]
            obj = self.serializer.concatenate([old_obj, new_obj])
            assert(len(obj) == 1)
            obj = obj[0]
            # Reset file so that header will be written
            self.reset_position(truncate=True)
        return super(FileComm, self).serialize(obj, **kwargs)
            
    def _send(self, msg):
        r"""Write message to a file.

        Args:
            msg (bytes, str): Data to write to the file.

        Returns:
            bool: Success or failure of writing to the file.

        """
        # Write header
        if msg != self.eof_msg:
            self.write_header()
        # Write message
        try:
            if msg != self.eof_msg:
                self.fd.write(msg)
                if self.append == 'ow':
                    self.fd.truncate()
            self.fd.flush()
        except (AttributeError, ValueError):  # pragma: debug
            if self.is_open:
                raise
            return False
        if msg != self.eof_msg and self.is_series:
            self.advance_in_series()
            self.debug("Advanced to %d", self._series_index)
        return True

    def _recv(self, timeout=0):
        r"""Reads message from a file.

        Args:
            timeout (float, optional): Time in seconds to wait for a message.
                Defaults to self.recv_timeout. Unused.

        Returns:
            tuple (bool, str): Success or failure of reading from the file and
                the read messages as bytes.

        """
        try:
            self.read_header()
            prev_pos = self.fd.tell()
            flag = True
            if self.read_meth == 'read':
                out = self.fd.read()
            elif self.read_meth == 'readline':
                out = self.fd.readline()
        except BaseException:  # pragma: debug
            # Use this to catch case where close called during receive.
            # In the future this should be handled via a lock.
            out = ''
        if len(out) == 0:
            if self.advance_in_series():
                self.debug("Advanced to %d", self._series_index)
                flag, out = self._recv()
            elif self.append and self.is_open:
                self.fd.seek(prev_pos)
                out = self.empty_bytes_msg
            else:
                out = self.eof_msg
        else:
            out = out.replace(self.platform_newline, self.serializer.newline)
            if flag and (not self.is_eof(out)):
                if (((self.read_meth == 'readline')
                     and out.startswith(self.serializer.comment))):
                    # Exclude comments
                    flag, out = self._recv()
                elif (((self.read_meth == 'read') and (prev_pos > 0)
                       and (not self.serializer.concats_as_str))):
                    # Rewind file and read entire contents if data was added to
                    # the file type using a serialization method that dosn't
                    # concatenate
                    self.reset_position()
                    flag, out = self._recv()
                elif (self.read_meth == 'read') and self.serializer.is_framed:
                    # Rewind if more than one frame read
                    len0 = len(out)
                    out = self.serializer.get_first_frame(out)
                    len1 = len(out)
                    if (len1 > 0) and (len0 != len1):
                        self.fd.seek(prev_pos + len1)
        return (flag, out)

    def purge(self):
        r"""Purge all messages from the comm."""
        if self.is_open and self.direction == 'recv':
            try:
                self.fd.seek(0, os.SEEK_END)
            except (AttributeError, ValueError):  # pragma: debug
                if self.is_open:
                    raise



import os
import tempfile
from yggdrasil.tests import assert_equal, assert_warns, assert_raises
from yggdrasil import config, backwards, tools
from yggdrasil.components import import_component


def make_temp(fname_base, count=1):
    r"""Create temporary copies of same file with different extensions."""
    fname_base = fname_base.lower()
    tempdir = os.path.normcase(os.path.normpath(tempfile.gettempdir()))
    if (tempdir + os.pathsep) not in os.environ['PATH']:
        os.environ['PATH'] = os.pathsep.join([tempdir, os.environ.get('PATH')])
    fname_pattern = fname_base + '.*'
    fname = os.path.join(tempdir, fname_base)
    out = []
    for i in range(count):
        fname_i = '%s.%d' % (fname, i)
        out.append(fname_i)
        if not os.path.isfile(fname_i):
            with open(fname_i, 'w') as fd:
                fd.write('Test file %d' % i)
    return tempdir, fname_pattern, out


def make_temp_single():
    r"""Create single temporary file."""
    return make_temp('single_test_file')


def make_temp_multiple():
    r"""Create multiple temporary files."""
    return make_temp('multiple_test_file', count=2)


def test_YggConfigParser():
    r"""Ensure that get returns proper defaults etc."""
    x = config.YggConfigParser()
    x.add_section('test_section')
    x.set('test_section', 'test_option', 'test_value')
    assert_raises(RuntimeError, x.update_file)
    assert_equal(x.file_to_update, None)
    assert_equal(x.get('test_section', 'test_option'), 'test_value')
    assert_equal(x.get('test_section', 'fake_option'), None)
    assert_equal(x.get('test_section', 'fake_option', 5), 5)
    assert_equal(x.get('fake_section', 'fake_option'), None)
    assert_equal(x.get('fake_section', 'fake_option', 5), 5)


def test_update_language_config():
    r"""Test updating configuration for installed languages."""
    drv = [import_component('model', l) for l in tools.get_supported_lang()]
    config.update_language_config(drv, overwrite=True, verbose=True)
    try:
        config.update_language_config(drv[0], disable_languages=[drv[0].language])
        config.update_language_config(drv[0],
                                      disable_languages=[drv[0].language],
                                      enable_languages=[drv[0].language])
    finally:
        config.update_language_config(drv[0], enable_languages=[drv[0].language])


def test_locate_file():
    r"""Test file location method."""
    # Missing file
    assert(not config.locate_file('missing_file.fake'))
    # Single file
    sdir, spat, sans = make_temp_single()
    sout = config.locate_file(spat)
    assert(isinstance(sout, backwards.string_types))
    assert_equal(sout, sans[0])
    # Multiple files
    mdir, mpat, mans = make_temp_multiple()
    with assert_warns(RuntimeWarning):
        mout = config.locate_file(mpat)
        assert(isinstance(mout, backwards.string_types))
        assert_equal(mout, mans[0])
    

def test_find_all():
    r"""Test find_all."""
    # Missing file
    assert(not config.find_all('missing_file.fake', 'invalid'))
    # Single file
    sdir, spat, sans = make_temp_single()
    sout = config.find_all(spat, sdir)
    assert(isinstance(sout, list))
    assert_equal(sout, sans)
    # Multiple files
    mdir, mpat, mans = make_temp_multiple()
    mout = config.find_all(mpat, mdir)
    assert(isinstance(mout, list))
    assert_equal(mout, mans)


# def test_update_config():
#     r"""Test update_config."""
#     test_cfg = os.path.join(tempfile.gettempdir(), 'test.cfg')
#     assert(not os.path.isfile(test_cfg))
#     if not tools.is_lang_installed('matlab'):  # pragma: no matlab
#         assert_warns(RuntimeWarning, config.update_config, test_cfg)
#     config.update_config(test_cfg)
#     assert(os.path.isfile(test_cfg))
#     os.remove(test_cfg)


def test_cfg_logging():
    r"""Test cfg_logging."""
    lvl = config.get_ygg_loglevel()
    config.set_ygg_loglevel(lvl)
    os.environ['YGG_SUBPROCESS'] = 'True'
    lvl = config.get_ygg_loglevel()
    config.set_ygg_loglevel(lvl)
    config.cfg_logging()
    del os.environ['YGG_SUBPROCESS']



from yggdrasil import backwards
from yggdrasil.serialize import format_message
from yggdrasil.serialize.SerializeBase import SerializeBase


class DirectSerialize(SerializeBase):
    r"""Class for directly serializing bytes."""

    _seritype = 'direct'
    _schema_subtype_description = ('Direct serialization of bytes.')
    default_datatype = {'type': 'bytes'}

    def func_serialize(self, args):
        r"""Serialize a message.

        Args:
            args: List of arguments to be formatted or numpy array to be
                serialized.

        Returns:
            bytes, str: Serialized message.

        """
        if (((self.extra_kwargs.get('format_str', None) is not None)
             and isinstance(args, list))):
            args = format_message(args, self.extra_kwargs['format_str'])
        return backwards.as_bytes(args)

    def func_deserialize(self, msg):
        r"""Deserialize a message.

        Args:
            msg: Message to be deserialized.

        Returns:
            obj: Deserialized message.

        """
        return msg

    @classmethod
    def concatenate(cls, objects, **kwargs):
        r"""Concatenate objects to get object that would be recieved if
        the concatenated serialization were deserialized.

        Args:
            objects (list): Objects to be concatenated.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            list: Set of objects that results from concatenating those provided.

        """
        return [b''.join(objects)]
    
    @classmethod
    def get_testing_options(cls):
        r"""Method to return a dictionary of testing options for this class.

        Returns:
            dict: Dictionary of variables to use for testing. Key/value pairs:
                kwargs (dict): Keyword arguments for comms tested with the
                    provided content.
                empty (object): Object produced from deserializing an empty
                    message.
                objects (list): List of objects to be serialized/deserialized.
                extra_kwargs (dict): Extra keyword arguments not used to
                    construct type definition.
                typedef (dict): Type definition resulting from the supplied
                    kwargs.
                dtype (np.dtype): Numpy data types that is consistent with the
                    determined type definition.

        """
        out = {'kwargs': {}, 'empty': b'', 'dtype': None,
               'typedef': cls.default_datatype,
               'extra_kwargs': {}}
        out['objects'] = [b'Test message\n', b'Test message 2\n']
        out['contents'] = b''.join(out['objects'])
        return out



import os
import argparse
ROOT_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                         '..', '..', '..'))
rj_include_dir0 = os.path.join(ROOT_PATH, 'yggdrasil', 'rapidjson', 'include')


def update_argparser(parser=None):
    r"""Update argument parser with language specific arguments.

    Args:
        parser (argparse.ArgumentParser, optional): Existing argument parser
            that should be updated. Default to None and a new argument parser
            will be created.

    Returns:
        argparse.ArgumentParser: Argument parser with language specific arguments.

    """
    if parser is None:
        parser = argparse.ArgumentParser("Run C installation script.")
    parser.add_argument('--rj-include-dir', '--rapidjson-include-dir',
                        default=rj_include_dir0,
                        nargs=1, help='Rapidjson include directory.')
    return parser


def install(args=None, rj_include_dir=None):
    r"""Check that rapidjson is installed.

    Args:
        args (argparse.Namespace, optional): Arguments parsed from the
            command line. Default to None and is created from sys.argv.
        rj_include_dir (str, optional): Full path to the include directory for
            the rapidjson header-only package. Defaults to None and args
            is checked for '--rj-include-dir' or '--rapidjson-include-dir'.

    Returns:
        bool: True if the installation is valid, False otherwise.

    """
    if args is None:
        args = update_argparser().parse_args()
    if rj_include_dir is None:
        rj_include_dir = args.rj_include_dir
    if not os.path.isdir(rj_include_dir):
        raise RuntimeError("RapidJSON sources could not be located. If you "
                           "cloned the git repository, initialize the rapidjson "
                           "git submodule by calling "
                           "'git submodule update --init --recursive' "
                           "from inside the repository.")
    if rj_include_dir != rj_include_dir0:
        def_config_file = os.path.join(ROOT_PATH, 'yggdrasil', 'defaults.cfg')
        try:
            import ConfigParser as configparser
        except ImportError:
            import configparser
        cfg = configparser.ConfigParser()
        cfg.read(def_config_file)
        if not cfg.has_section('c'):
            cfg.add_section('c')
        cfg.set('c', 'rapidjson_include', rj_include_dir)
        with open(def_config_file, 'w') as fd:
            cfg.write(fd)
    return True


if __name__ == "__main__":
    out = install()
    if out:
        print("rapidjson installed.")
    else:
        raise Exception("Failed to located rapidjson.")



from yggdrasil.metaschema.datatypes import MetaschemaTypeError
from yggdrasil.metaschema.properties.MetaschemaProperty import (
    MetaschemaProperty)
from yggdrasil.metaschema.properties.JSONObjectMetaschemaProperties import (
    PropertiesMetaschemaProperty)


class ArgsMetaschemaProperty(MetaschemaProperty):
    r"""Property class for 'args' property."""

    name = 'args'
    schema = {'description': ('Arguments required to recreate a class instance.'),
              'type': 'object'}
    _instance_dict_attr = ['input_arguments', 'input_args']

    @classmethod
    def instance2args(cls, instance):
        r"""Get input arguments from a class instance.

        Args:
            instance (object): Instance of a Python class.

        Returns:
            dict: Input arguments for re-creating the instance.

        """
        for k in cls._instance_dict_attr:
            if hasattr(instance, k):
                return getattr(instance, k)
            elif hasattr(instance, 'get_' + k):
                return getattr(instance, 'get_' + k)()
            elif hasattr(instance, '_' + k):
                return getattr(instance, '_' + k)
        raise MetaschemaTypeError('Could not locate dictionary of arguments.')

    @classmethod
    def encode(cls, instance, typedef=None):
        r"""Encoder for the 'args' property."""
        typedef_args = None
        # if isinstance(typedef, dict) and ('args' in typedef):
        #     typedef_args = typedef['args']
        args = cls.instance2args(instance)
        return PropertiesMetaschemaProperty.encode(args, typedef_args)

    @classmethod
    def compare(cls, *args, **kwargs):
        r"""Comparison method for 'args' container property."""
        for e in PropertiesMetaschemaProperty.compare(*args, **kwargs):
            yield e



import numpy as np
from yggdrasil.interface.YggInterface import YggRpcServer


def get_fibonacci(n):
    r"""Compute the nth number of the Fibonacci sequence.

    Args:
        n (int): Index of the Fibonacci number in the Fibonacci sequence that
            should be returned.

    Returns:
        int: The nth Fibonacci number.

    """
    pprev = 0
    prev = 1
    result = 1
    fib_no = 1
    while fib_no < n:
        result = prev + pprev
        pprev = prev
        prev = result
        fib_no = fib_no + 1
    return result


def main():
    r"""Function to execute server communication and computation in a loop."""

    print('Hello from Python server!')

    # Create server-side rpc conneciton using model name
    rpc = YggRpcServer("server", "%d", "%d")

    # Continue receiving requests until the connection is closed when all
    # clients have disconnected.
    while True:
        print('server: receiving...')
        retval, rpc_in = rpc.recv()
        if not retval:
            print('server: end of input')
            break

        # Compute fibonacci number
        n = rpc_in[0]
        print('server: Received request for Fibonacci number %d' % n)
        result = get_fibonacci(n)
        print('server: Sending response for Fibonacci number %d: %d' % (n, result))

        # Send response back
        flag = rpc.send(np.int32(result))
        if not flag:
            raise RuntimeError('server: ERROR sending')

    print('Goodbye from Python server')

    
if __name__ == '__main__':
    main()



import os
from yggdrasil.languages import get_language_dir
from yggdrasil.drivers.PythonModelDriver import PythonModelDriver


_model_script = os.path.join(get_language_dir('lpy'), 'lpy_model.py')


class LPyModelDriver(PythonModelDriver):  # pragma: lpy
    r"""Class for running LPy models."""

    _schema_subtype_description = ('Model is an LPy system.')
    
    language = 'lpy'
    language_ext = '.lpy'
    # base_languages = ['python']  # Uncomment if PythonModelDriver not parent
    default_interpreter_flags = [_model_script]
    interface_dependencies = ['openalea.lpy']
    function_param = None

    @classmethod
    def language_version(cls, **kwargs):
        r"""Determine the version of this language.

        Args:
            **kwargs: Keyword arguments are passed to cls.run_executable.

        Returns:
            str: Version of compiler/interpreter for this language.

        """
        try:
            import openalea.lpy
            return openalea.lpy.__version__.LPY_VERSION_STR
        except ImportError:  # pragma: debug
            raise RuntimeError("openalea.lpy not installed.")
